program ApplyWgt3D
! Output the data space weighting file
! Dimension: n1 = nrx
!            n2 = nry
!            n3 = nt


  use sep
  
  implicit none
  
  integer                   :: nrx,nry,nt
  integer                   :: irx,iry,it
  real                      :: orx, ory, ot
  real                      :: drx, dry, dt
  character(len=128)        :: outfile_, Wgtfile_
  real, dimension(:,:,:), allocatable   ::  inArray, outArray, WgtArray

  call sep_init()

  call from_param("outfile", outfile_);
  call from_param("Wgtfile", Wgtfile_);


  call from_history("n1",nrx);     call from_history("d1",drx);    call from_history("o1",orx);
  call from_history("n2", nry);     call from_history("d2",dry);    call from_history("o2",ory);
  call from_history("n3", nt);     call from_history("d3",dt);    call from_history("o3",ot);

  allocate(inArray(nrx,nry,nt), outArray(nrx,nry,nt), WgtArray(nrx,nry,nt))


  write(0,*) "***** ApplyWgt3D ************"
  write(0,*) 'nrx, drx, orx', nrx, drx, orx
  write(0,*) 'nrx, drx, orx', nry, dry, ory
  write(0,*) 'nt, dt, ot', nt, dt ,ot
  write(0,*) "*****************************"

  write(0,*) "reading input file"
  call sep_read(inArray);

  write(0,*) "reading weight file"
  call sep_read(WgtArray, Wgtfile_);

  write(0,*) "applying weights"
  do it=1,nt
  do iry=1,nry
  do irx=1,nrx
        outArray(irx,iry,it) = inArray(irx,iry,it)*WgtArray(irx,iry,it)
    end do
  end do
  end do

  write(0,*) "outputting"
  call sep_write(outArray,outfile_,dim1=nrx, dim2=nry, dim3=nt, esize=4)

         call auxputch("n1",'i',nrx, trim(outfile_));     
         call auxputch("d1",'f',drx, trim(outfile_));   
         call auxputch("o1",'f',orx,trim(outfile_));
         call auxputch("label1",'s','Rx',trim(outfile_));

         call auxputch("n2",'i',nry, trim(outfile_));     
         call auxputch("d2",'f',dry, trim(outfile_));   
         call auxputch("o2",'f',ory,trim(outfile_));
         call auxputch("label2",'s','Ry',trim(outfile_));

	  call auxputch("n3",'i',nt, trim(outfile_));     
	  call auxputch("d3",'f',dt, trim(outfile_));   
	  call auxputch("o3",'f',ot,trim(outfile_));
          call auxputch("label3",'s','time',trim(outfile_));


end program
