program DataWgt3d
! Output the data space weighting file
! Dimension: n1 = nrx
!            n2 = nt
!            n3 = UD


  use sep
  
  implicit none
  
  integer                   :: nrx, nry, nt
  integer                   :: irx, iry, it
  integer                   :: imid, itb, ite, wd, ud, mode
  integer                   :: iitb, iite
  real                      :: orx, ory, ot, zw
  real                      :: drx, dry, dt, pow, t
  real                      :: Dep , hx, hy, vel, Gx, Gy, Gz
  real                      :: xx, yy, tmute
  real                      :: endvalX, endvalY
  real                      :: cent1, cent2, A1, A2, dist, amp, pi=3.14159265359
  character(len=128)        :: outfile_, Wgtfile_
  real,  dimension(:,:), allocatable  :: tape
  real, dimension(:,:,:), allocatable   ::  inArray, outArray, WgtArray

  call sep_init()

  call from_param("outfile", outfile_);
  call from_param("Wgtfile", Wgtfile_);
  call from_param("UD", UD);  ! 1=up , 2=down
  call from_param("mode", mode, 1)   ! 1=output Data weighting only; 2=apply wgt as well

  call from_param("pow",pow,1.)
  call from_param("Gx",Gx);  call from_param("Gy",Gy);   call from_param("Gz",Gz);
  call from_param("A1", A1, 0.4);    call from_param("A2", A2, 0.05);
  call from_param("zw",zw)
  call from_param("wd",wd,500);  call from_param("tmute",tmute,2.0);

  call from_history("n1",nrx);     call from_history("d1",drx);    call from_history("o1",orx);
  call from_history("n2", nry);     call from_history("d2",dry);    call from_history("o2",ory);
  call from_history("n3", nt);     call from_history("d3",dt);    call from_history("o3",ot);

  endvalX = orx+(nrx-1)*drx;  endvalY = ory+(nry-1)*dry;

  if ((A1 .lt.  0.) .or. (A1 .gt. 1.)) call erexit('error for A1 value here')
  if ((A2 .lt.  0.) .or. (A2 .gt. 1.)) call erexit('error for A2 value here')
  
  allocate(tape(nrx,nry))
  allocate(inArray(nrx,nry,nt), outArray(nrx,nry,nt), WgtArray(nrx,nry,nt))


  vel = 1500.
  cent1 = 1.0 - A1

  if (UD == 1) then
	Dep = 900.
  else 
        Dep = 3*900.
  end if



  write(0,*) '*****************************'
  write(0,*) 'nrx, drx, orx', nrx, drx, orx
  write(0,*) 'nry, dry, ory', nry, dry, ory
  write(0,*) 'nt, dt, ot', nt, dt ,ot
  write(0,*) 'wd, vel, cent1, A1', wd, vel, cent1, A1
  write(0,*) 'Dep, Gx, Gy', Dep, Gx, Gy
  write(0,*) 'orx+(nrx-1)*drx', endvalX, orx+(nrx-1)*drx
  write(0,*) 'ory+(nry-1)*dry', endvalY, ory+(nry-1)*dry
  write(0,*) 'pow', pow
   write(0,*) '*****************************'

  do iry=1,nry
	yy = ((iry-1)*dry+ory)
       hy = Gy - yy
  do irx=1,nrx     
	xx = ((irx-1)*drx+orx)
       hx = Gx - xx
       if (UD == 1) then
         tape(irx,iry) = sqrt(Gz**2 + hx**2 + hy**2)/vel + tmute
       else
	 tape(irx,iry) = sqrt(Dep**2 + hx**2 + hy**2)/vel + tmute
       end if

	if ((mod(irx,5) == 1) .and. (iry == 45) ) write(0,*) 'xx,yy,tt',irx,yy,tape(irx,iry)

  end do
  end do


  write(0,*) 'pass 2'


  if (UD == 1) then
   write(0,*) 'constructing Up-going weights'
! Setting up the up-going Data weighting
   do iry=1,nry
 !     write(0,*) 'iry,nry', iry, nry
   do irx=1,nrx
      imid = (tape(irx,iry) - ot)/dt + 1.5
      itb = imid-wd/2
      ite = itb + wd  
     if ((mod(irx,5) == 1) .and. (iry == 45) ) write(0,*) 'tape, itb,ite, imid', tape(irx,iry),itb,ite, imid
      do it=max(itb,1+1),min(ite,nt-1)
          dist = wd-(itb-it)
         ! Cosine Taper
          WgtArray(irx,iry,it) = (cent1 + A1*cos(pi*(dist)/wd))
      end do
	 iitb = min(max(itb,2),nt) 
         amp = WgtArray(irx,iry,iitb)
         WgtArray(irx,iry,1:iitb-1) = amp
         iite = max(min(ite,nt-1),1)
         amp = WgtArray(irx,iry,iite)
	 WgtArray(irx,iry,iite+1:nt) = amp
  !    do it=iite,nt
  !       WgtArray(irx,iry,iite+1:nt) = amp*(1+(it-iite)/3000.)**pow
  !    end do

   end do
   end do

 else if (UD == 2) then
   write(0,*) 'constructing Dn-going weights'

  cent2 = 1.0-A2

! Setting up the down-going Data weighting
   do iry=1,nry
   do irx=1,nrx
      imid = (tape(irx,iry) - ot)/dt + 1.5
      itb = imid-wd/2
      ite = itb + wd     

 !     if (irx .gt. 220) write(0,*) 'pass 3.5', irx, nrx, ite, itb    
      do it=max(itb,1),min(ite,nt)
          dist = wd-(itb-it)
         ! Cosine Taper
          WgtArray(irx,iry,it) = (cent2 - A2*cos(pi*(dist)/wd))
      end do

	 iitb = min(max(itb,2),nt) 
         iite = max(min(ite,nt-1),1)
 !      if (irx .gt. 220) write(0,*) 'pass 3.75',ite, itb, iite, iitb

         amp = WgtArray(irx,iry,iitb)
         WgtArray(irx,iry,1:iitb-1) = amp


         amp = WgtArray(irx,iry,iite)
         WgtArray(irx,iry,iite+1:nt) = amp
   end do
   end do

  end if


           write(0,*) "pass 3"



  ! Apply a power factor to the weight
  ! weight = weight * (t)**pow
  if (pow .gt. 0.) then
   write(0,*) "gain data by a factor t**pow", pow
  do iry=1,nry
  do irx=1,nrx
    do it=1,nt
       t = (it-1)*dt+ot
  	if ((irx .eq. nrx/2) .and.   (mod(it,2000) .eq. 0)) write(0,*) 't**pow', t**pow  
        WgtArray(irx,iry,it) = WgtArray(irx,iry,it)*((t-tape(irx,iry)+2)/(8.0))**pow
    end do     
  end do
  end do
  end if

  !Apply muting to area where there is no signal
  do iry=1,nry
  do irx=1,nrx
      imid = (tape(irx,iry) - ot)/dt + 1.5
      itb = imid-wd/2  
      iitb = min(max(itb,2),nt) 

         amp = WgtArray(irx,iry,iitb)
         WgtArray(irx,iry,1:iitb-1) = 0.0
     
  end do
  end do


  write(0,*) 'pass 5'

  if (mode .eq. 2) then  ! also apply weight to the input file

    call sep_read(inArray);

    write(0,*) 'pass 6'

    do iry=1,nry
    do irx=1,nrx
      do it=1,nt
          outArray(irx,iry,it) = inArray(irx,iry,it)*WgtArray(irx,iry,it)
      end do
    end do
    end do


  call sep_write(outArray,outfile_,dim1=nrx, dim2=nry, dim3=nt, esize=4)

         call auxputch("n1",'i',nrx, trim(outfile_));     
         call auxputch("d1",'f',drx, trim(outfile_));   
         call auxputch("o1",'f',orx,trim(outfile_));
         call auxputch("label1",'s','Rx',trim(outfile_));

         call auxputch("n2",'i',nry, trim(outfile_));     
         call auxputch("d2",'f',dry, trim(outfile_));   
         call auxputch("o2",'f',ory,trim(outfile_));
         call auxputch("label2",'s','Ry',trim(outfile_));

	  call auxputch("n3",'i',nt, trim(outfile_));     
	  call auxputch("d3",'f',dt, trim(outfile_));   
	  call auxputch("o3",'f',ot,trim(outfile_));
          call auxputch("label3",'s','time',trim(outfile_));

   end if

!-----------------------------------------------------------------------------
         call auxputch("n1",'i',nrx, trim(Wgtfile_));     
         call auxputch("d1",'f',drx, trim(Wgtfile_));   
         call auxputch("o1",'f',orx,trim(Wgtfile_));
         call auxputch("label1",'s','Rx',trim(Wgtfile_));

         call auxputch("n2",'i',nry, trim(Wgtfile_));     
         call auxputch("d2",'f',dry, trim(Wgtfile_));   
         call auxputch("o2",'f',ory,trim(Wgtfile_));
         call auxputch("label2",'s','Ry',trim(Wgtfile_));

	  call auxputch("n3",'i',nt, trim(Wgtfile_));     
	  call auxputch("d3",'f',dt, trim(Wgtfile_));   
	  call auxputch("o3",'f',ot,trim(Wgtfile_));
          call auxputch("label3",'s','time',trim(Wgtfile_));

  call sep_write(WgtArray,Wgtfile_,dim1=nrx, dim2=nry, dim3=nt, esize=4)

end program
