program Forward3D_Off
! takes an ADCIG and do forward modeling 

  use sep
  use omp_lib
  use rtm3d_operAngle
  use ADCIG_mod
  use waveprop3d_mod
  use type3d_mod
  use smooth3d_mod
 ! use ifport

  implicit none

  type(velo) :: vel
  type(sou_grid)  :: sougrid
  type(rec_grid)  :: recgrid
  type(bc_values) :: bc


  logical(4) :: success        
  logical                   :: smooth, verb, wfldout
  integer                   :: it, nt, nw, sstat, jsnap
  integer                   :: iz, nz, bndz, nnz, nmz
  integer                   :: ix, nx, bndx, nnx, nmx
  integer                   :: iy, ny, bndy, nny, nmy
  integer                   :: ihx, nhx, iax, nax
  real                      :: dax, oax, dhx, ohx
  integer                   :: ii
  character(len=128)        :: refl_, velout_, wfldsou_, wfldrec_, wfldmod_, out_
  character(len=128)        :: vp_, MWfile_, DWfile_, swin_
  real                      :: rho
  real, dimension(:,:,:), allocatable  :: fs1, data
  real, dimension(:,:,:,:), allocatable  :: model

   ! OMP stuff 
  integer :: inode,node,nodes, Totalnodes

  ! Initial timer
  integer*4                 :: now2(3), now1(3)
  call itime(now1)


  call sep_init()


  
  
  !!======================================================
  !! Parameter reading

   call from_param("verb",verb,.false.);  call from_param("smooth",smooth,.false.);
   call from_param("wfrout",bc%wfrout,.false.);
   call from_param("wfmout",bc%wfmout,.false.);
   call from_param("lpfil", bc%lpfil, .false.);
   call from_param("VDer", bc%VDer, .false.);
   call from_param("DW", bc%DW, .false.);              
   call from_param("gauss", bc%gauss, .false.);


   ! Read file names
   call from_param("velin", vp_);      ! migration velocity
   call from_param("wfldsou", wfldsou_);     ! pre-computed source-side wavefield
   call from_param("bfile", sougrid%bfile, 0);   ! wfldsou file indices 
   call from_param("efile",sougrid%efile, 0);
   call from_param("reflfile",refl_)      ! reflectivity model
   call from_param("MWfile",MWfile_);     ! Model weighting function

   ! Read back ground velocity param
   call from_aux(vp_, "n1", vel%z%n);  call from_aux( vp_, "d1", vel%z%d);  call from_aux( vp_, "o1",vel%z%o);
   call from_aux( vp_, "n2", vel%x%n);  call from_aux( vp_, "d2", vel%x%d);  call from_aux( vp_, "o2",vel%x%o);
   call from_aux( vp_, "n3", vel%y%n);  call from_aux( vp_, "d3", vel%y%d);  call from_aux( vp_, "o3",vel%y%o);
   call from_param('zw', vel%zw);
   nw = (vel%zw-vel%z%o)/vel%z%d + 0.5
   call from_aux(refl_,"n1", nmz);  call from_aux(refl_,"n2",nmx); call from_aux(refl_,"n3",nmy);
   


   !All shots has the same source wavelet
   call getfilename(wfldsou_,swin_,sougrid%bfile)
   call from_aux(swin_, "n4", sougrid%ntsav);  call from_aux(swin_,"d4", sougrid%dtsav);   call from_aux(swin_,"o4", sougrid%otsav);
   
   
   call from_param('nrx', recgrid%x%n);    call from_param('drx',recgrid%x%d );     call from_param('orx', recgrid%x%o);
   call from_param('nry', recgrid%y%n);    call from_param('dry',recgrid%y%d );     call from_param('ory', recgrid%y%o);
   call from_param('nt', recgrid%nt);     
   call from_param('dt', recgrid%dt);      
   call from_param('ot', recgrid%ot);
   call from_param('nrt', recgrid%t%n, recgrid%nt);     
   call from_param('drt', recgrid%t%d, recgrid%dt);      
   call from_param('ort', recgrid%t%o,recgrid%ot);
   call from_param('zr', recgrid%z%o);      
   
   

   ! Read in Density
   call from_param('rho', rho);

   
   !! taper coefficients and padding
   call from_param("xtapco",bc%xtapco,0.0002);     call from_param("ytapco",bc%ytapco,0.0002);  
   call from_param("ztapco",bc%ztapco,0.0002)
   call from_param("nop",bc%nop,8)
   call from_param("bndz",bc%bndz,bc%nop); if(bc%bndz<bc%nop) bc%bndz=bc%nop; 
   call from_param("bndx",bc%bndx,bc%nop); if(bc%bndx<bc%nop) bc%bndx=bc%nop;
   call from_param("bndy",bc%bndy,bc%nop); if(bc%bndy<bc%nop) bc%bndy=bc%nop;

   ! Data weighting and Masking
   call from_param("bmaskX",bc%bmaskX,recgrid%x%o);  
   call from_param("emaskX",bc%emaskX,recgrid%x%o+(recgrid%x%n-1)*recgrid%x%d);
   call from_param("bmaskY",bc%bmaskY,recgrid%y%o);  
   call from_param("emaskY",bc%emaskY,recgrid%y%o+(recgrid%y%n-1)*recgrid%y%d);
   if (bc%DW) then  
  	 call from_param("DWfile",DWfile_); 
   end if

   !! Auxillary output files
   call from_param("outtag", out_);
   call from_param("velout", velout_, "velfile.H");
   call from_param("wfldrec", wfldrec_, "wfldrec.H");
   call from_param("wfldmod", wfldmod_, "wfldmod.H");
   call from_Param("jsnap", jsnap, 50);

   ! Parallel Programming
   call from_param("nodes" ,nodes, 8)
   Totalnodes =omp_get_num_procs()
   call omp_set_num_threads(nodes); 
    write(0,*) 'OMP: using', nodes, 'out of', Totalnodes
    
  !! Set temporary binary datapath
 ! success = setenvqq("DATAPATH=/tmp/")

   ! Error check

   if (jsnap .le. 0) jsnap=50
  
!   write(0,*) 'sougrid dt', sougrid%t%n, sougrid%t%d
!   write(0,*) 'recgrid dt', recgrid%t%n, recgrid%t%d	
!   if (sougrid%t%n .gt. recgrid%t%n) call erexit('sougrid dt not equal to recgrid dt')

   if (vel%z%n .ne. (nw+nmz)) call erexit('vel nz not equal to model nz + nw')
   if (vel%x%n .ne. (nmx)) call erexit('vel nx not equal to model nx')  
   if (vel%y%n .ne. (nmy)) call erexit('vel ny not equal to model ny')  
   if (recgrid%t%n .gt. 6000) call erexit('too many time steps, nt < 6000')
 
   if ((recgrid%z%n .eq. 2) .and. (recgrid%z%d .eq. 0)) call erexit('Expect over/under but found dz=0')
   
   if (bc%wfrout) then
      if (recgrid%x%n*recgrid%y%n*vel%z%n*nt/jsnap*4 .gt. 1.e10) call erexit('size of wfr is too big, please increase jsnap')
   end if

   !! Definition
   vel%smooth = smooth
   sougrid%x%n = 1
   sougrid%y%n = 1
   nnz = vel%z%n;   nnx = vel%x%n;    nny = vel%y%n;
   bndx = bc%bndx;  bndy = bc%bndy;   bndz = bc%bndz;
   vel%z%n = vel%z%n + 2*bndz; 
   vel%x%n = vel%x%n+2*bndx
   vel%y%n = vel%y%n+2*bndy
   nz = vel%z%n;   nx = vel%x%n;   ny = vel%y%n;
   nt=recgrid%t%n
   nw = (vel%zw-vel%z%o)/vel%z%d + 0.5

  
  !! hard-coded for now
   nhx=30
   dhx = vel%x%d
   ohx = -nhx*dhx
   
   nax = 60.
   dax = 0.06
   oax = -1.8

   !*********************************
   write(0,*) '**************************'
   write(0,*) 'vel: nz, dz, oz ', nz, vel%z%d, vel%z%o
   write(0,*) 'vel: nx, dx, ox',  nx, vel%x%d, vel%x%o
   write(0,*) 'vel: ny, dy, oy',  ny, vel%y%d, vel%y%o
   write(0,*) 'vel: zw, nw'         ,  vel%zw, nw
   write(0,*) 'sou: nx, dx, ox',  sougrid%x%n, sougrid%x%d, sougrid%x%o  
   write(0,*) 'sou: ny, dy, oy',  sougrid%y%n, sougrid%y%d, sougrid%y%o  
   write(0,*) 'sou: bfile, efile', sougrid%bfile, sougrid%efile
   write(0,*) 'sou: ntsav, dtsav, otsav',     sougrid%ntsav, sougrid%dtsav, sougrid%otsav
   write(0,*) 'rec: nx, dx, ox', recgrid%x%n, recgrid%x%d, recgrid%x%o
   write(0,*) 'rec: ny, dy, oy', recgrid%y%n, recgrid%y%d, recgrid%y%o
   write(0,*) 'rec: nrt, drt, ort', recgrid%t%n, recgrid%t%d, recgrid%t%o
   write(0,*) 'mod: nt, dt, ot', recgrid%nt, recgrid%dt, recgrid%ot
   write(0,*) 'hx,  nhx, dhx, ohx', nhx, dhx, ohx
   write(0,*) 'ax, nax, dax, oax', nax, dax, oax
   write(0,*) 'smooth, wfrout', smooth, bc%wfrout
   write(0,*) 'orz, drz', recgrid%z%o, recgrid%z%d
   write(0,*) 'bndz, bndy, nop', bc%bndz, bc%bndx, bc%bndy, bc%nop
   write(0,*) 'xtapco, ztapco', bc%xtapco, bc%ytapco, bc%ztapco
   write(0,*) 'jsnap', jsnap
   write(0,*) 'lpfil, VDer,guass', bc%lpfil, bc%VDer, bc%gauss
   write(0,*) 'DW, bmask, emask', bc%DW, bc%bmaskX, bc%emaskX, bc%bmaskY, bc%emaskY
   write(0,*) '**************************'

   !----------------------------------------------------------------

   ! Output file
   
  
   call auxputch("n1",'i', recgrid%x%n, trim(out_));     call auxputch("label1",'s','x', trim(out_));
   call auxputch("d1",'f', recgrid%x%d, trim(out_));     call auxputch("o1",'f',recgrid%x%o, trim(out_)); 

   call auxputch("n2",'i', recgrid%y%n, trim(out_));     call auxputch("label2",'s','y', trim(out_));
   call auxputch("d2",'f', recgrid%y%d, trim(out_));     call auxputch("o2",'f',recgrid%y%o, trim(out_));

   call auxputch("n3",'i',  nt, trim(out_));             call auxputch("label3",'s',  't', trim(out_));
   call auxputch("d3",'f', recgrid%t%d, trim(out_));     call auxputch("o3",'f', recgrid%t%o, trim(out_));

   
   

   
   ! Allocation
   write(0,*) '*** allocating data ***', recgrid%x%n*recgrid%y%n*recgrid%t%n
   allocate(data(recgrid%x%n,recgrid%y%n,recgrid%t%n))
      write(0,*) '*** allocating model ***', nnx,nny,(nnz-nw),nnx*nny*(nnz-nw)
   allocate(model((nnz-nw),nnx,nny,1+2*nhx))


   ! initialize
   do it = 1,recgrid%t%n
     do iy=1,recgrid%y%n
       do ix=1,recgrid%x%n
         data(ix,iy,it) = 0.;
       end do
     end do
   end do
   do iy = 1,nny
     do ix=1,nnx
       do iz=1,nnz-nw
         model(iz,ix,iy,:) = 0.;
       end do
     end do
   end do

   write(0,*) 'before reading refl model', refl_
  

   ! Read in reflectivity model        
    call sep_read(model, refl_);

write(0,*) 'pass 3'



   ! Call L_initiate
     sstat=L_init(sougrid, recgrid, vel, nhx, wfldsou_,vp_,rho, bc, wfldrec_in=wfldrec_, & 
     wfldmod_in=wfldmod_, verb_in=verb,jsnap_in=jsnap, MWfile_in=MWfile_, DWfile_in=DWfile_,nodes_in=nodes)
           



   !! Forward Model: reflectivity ----> data 
   !! adj = false
   sstat = L_oper(.false.,.false.,model, data)

   write(0,*) 'write date to out_', sum(data)

   call sep_write(data, out_)


   if (bc%wfrout) then
         call auxputch("n1",'i',vel%z%n,trim(wfldrec_));   
         call auxputch("d1",'f',vel%z%d,trim(wfldrec_));   call auxputch("o1",'f',vel%z%o,trim(wfldrec_));
	 call auxputch("label1", 's', 'Z', trim(wfldrec_));

         call auxputch("n2",'i',recgrid%x%n+2*bndx,trim(wfldrec_));   
         call auxputch("d2",'f',recgrid%x%d,trim(wfldrec_));   
         call auxputch("o2",'f',recgrid%x%o+(-bndx-1)*recgrid%x%d,trim(wfldrec_)); 
         call auxputch("label2", 's', 'X', trim(wfldrec_));


         call auxputch("n3",'i',recgrid%y%n+2*bndy,trim(wfldrec_));   
         call auxputch("d3",'f',recgrid%y%d,trim(wfldrec_));   
         call auxputch("o3",'f',recgrid%y%o+(-bndy-1)*recgrid%y%d,trim(wfldrec_));
	 call auxputch("label3", 's', 'Y', trim(wfldrec_));

         call auxputch("n3",'i',recgrid%t%n/jsnap,trim(wfldrec_));   
         call auxputch("d3",'f',recgrid%t%d*jsnap,trim(wfldrec_));   
	 call auxputch("o3",'f',recgrid%t%o+(jsnap-1)*recgrid%t%d*jsnap,trim(wfldrec_));
 	 call auxputch("label3", 's', 'time', trim(wfldrec_));
   end if



   call sep_close()

     
   deallocate(data, model);


   ! Stop timer
   call itime(now2)
   write(0,*) 'compute time Forward3D (s)', 3600*(now2(1)-now1(1))+60.*(now2(2)-now1(2))+(now2(3)-now1(3))


end program
