program Mig_Main3D_Off
  ! Prestack Reverse Time Migration (with angle domain )
  ! Input: Expect one set of data 
  !      : axis 1=x, 2=y 3=t
  !      : Expect Sou wavefield to be pre-computed
  !      : axis 1=nz, 2=nx, 3=ny, 4=nt,     
  ! Propagation Grid: use the mesh defined by velocity model plus padding bndx, bndy, bndz 

  

  use sep
  use omp_lib
  use rtm3d_operAngle
  use ADCIG_mod
  use waveprop3d_mod
  use type3d_mod
  use smooth3d_mod
  use ifport

  implicit none

  type(velo) :: vel
  type(sou_grid)  :: sougrid
  type(rec_grid)  :: recgrid
  type(bc_values) :: bc

  logical(4) :: success      
  logical                   :: smooth, verb
  integer                   :: it, nt, nw, id, sstat
  integer                   :: iz, nz, nnz, bndz
  integer                   :: ix, nx, nnx, bndx
  integer                   :: iy, ny, nny, bndy
  integer                   :: ihx, nhx, iax, nax
  real                      :: dax, oax, dhx, ohx
  integer                   :: jsnap, nnrt
  character(len=128)        :: rec_, velout_, wfldsou_, wfldrec_, wfldmod_
  character(len=128)        :: vp_, out_, out2_, MWfile_, DWfile_, swin_
  real                      :: rho, tmax
  real, dimension(:,:,:), allocatable  :: data
  real, dimension(:,:,:,:), allocatable  :: model

  ! OMP stuff 
  integer :: inode,node,nodes,Totalnodes

  ! Initial timer
  integer*4                 :: now2(3), now1(3)
  call itime(now1)


  call sep_init()


  !!======================================================
  !! Parameter reading
  
  write(0,*) ' This is the beginning of LsqMig_main'


   call from_param("verb",verb,.false.);     call from_param("smooth",smooth,.false.);   
   call from_param("wfmout",bc%wfmout,.false.);
   call from_param("wfrout",bc%wfrout,.false.);
   call from_param("lpfil", bc%lpfil, .false.);
   call from_param("VDer", bc%VDer, .false.);         
   call from_param("DW", bc%DW, .false.);
   call from_param("gauss", bc%gauss, .false.);


write(0,*) 'Input 1'
if (1 == 1) then   

   ! Read file name
   call from_param("velin", vp_);         ! migration velocity
   call from_param("wfldsou", wfldsou_);  ! pre-computed source-side wavefield
   call from_param("bfile", sougrid%bfile, 0);   call from_param("efile",sougrid%efile, 0);
   call from_param("recfile",rec_);       ! rec-data (to be back propagated)
   call from_param("MWfile",MWfile_);     ! Model weighting function


   ! Read back ground velocity param
   call from_aux(vp_, "n1", vel%z%n);  call from_aux( vp_, "d1", vel%z%d);  call from_aux( vp_, "o1",vel%z%o);
   call from_aux( vp_, "n2", vel%x%n);  call from_aux( vp_, "d2", vel%x%d);  call from_aux( vp_, "o2",vel%x%o);
   call from_aux( vp_, "n3", vel%y%n);  call from_aux( vp_, "d3", vel%y%d);  call from_aux( vp_, "o3",vel%y%o);
   call from_param('zw',vel%zw);
    nw = (vel%zw-vel%z%o)/vel%z%d + 0.5



   !All shots has the same source wavelet
   call getfilename(wfldsou_,swin_,sougrid%bfile)
   call from_aux(swin_, "n4", sougrid%ntsav);  call from_aux(swin_,"d4", sougrid%dtsav);   call from_aux(swin_,"o4", sougrid%otsav);


   ! Read rec grid param
   call from_aux(rec_, "n1", recgrid%x%n);  call from_aux(rec_,"d1", recgrid%x%d);   call from_aux(rec_,"o1", recgrid%x%o);
   call from_aux(rec_, "n2", recgrid%y%n);  call from_aux(rec_,"d2", recgrid%y%d);   call from_aux(rec_,"o2", recgrid%y%o);
   call from_aux(rec_, "n3", recgrid%t%n);  call from_aux(rec_,"d3", recgrid%t%d);   call from_aux(rec_,"o3", recgrid%t%o);
   call from_param('zr', recgrid%z%o);

   ! Modeling grid time param
   call from_param('nt', recgrid%nt);     
   call from_param('dt', recgrid%dt);      
   call from_param('ot', recgrid%ot);

   ! Read in Density
   call from_param('rho', rho);
  
   
   !!  taper coefficients and padding
   call from_param("xtapco",bc%xtapco,0.0002); call from_param("ytapco",bc%ytapco,0.0002);
   call from_param("ztapco",bc%ztapco,0.0002)
   call from_param("nop",bc%nop,8)
   call from_param("bndz",bc%bndz,bc%nop); if(bc%bndz<bc%nop) bc%bndz=bc%nop; 
   call from_param("bndx",bc%bndx,bc%nop); if(bc%bndx<bc%nop) bc%bndx=bc%nop;
   call from_param("bndy",bc%bndy,bc%nop); if(bc%bndy<bc%nop) bc%bndy=bc%nop;


   ! Data weighting and Masking
   call from_param("bmaskX",bc%bmaskX,recgrid%x%o);  
   call from_param("emaskX",bc%emaskX,recgrid%x%o+(recgrid%x%n-1)*recgrid%x%d);
   call from_param("bmaskY",bc%bmaskY,recgrid%y%o);  
   call from_param("emaskY",bc%emaskY,recgrid%y%o+(recgrid%y%n-1)*recgrid%y%d);
   if (bc%DW) then  
  	 call from_param("DWfile",DWfile_); 
   end if

   !! Auxillary output files
   call from_param("outtag", out_);
!   call from_param("outtag2", out2_);
   call from_param("velfile", velout_, "velfile.H");
   call from_param("wfldrec", wfldrec_, "wfldrec.H");
   call from_param("wfldmod", wfldmod_, "wfldmod.H");
   call from_Param("jsnap", jsnap, 50);


   ! Parallel Programming
   call from_param("nodes" ,nodes, 8)
   !nodes=8
   Totalnodes =omp_get_num_procs()
   call omp_set_num_threads(nodes); 
    write(0,*) 'OMP: using', nodes , 'out of', Totalnodes
    

  !! Set temporary binary datapath
!  success = setenvqq("DATAPATH=/tmp/")

   ! Error check


   if (jsnap .le. 0) jsnap=50
!   if (recgrid%x%d .ne. vel%x%d) call erexit('rec dx does not match vel dx')
!   if (recgrid%y%d .ne. vel%y%d) call erexit('rec dy does not match vel dy')
!   if (sougrid%t%n .gt. recgrid%t%n) call erexit('nst should be less than nrt') 
   
   if (recgrid%t%n .gt. 6001) call erexit('too many time steps, nt < 5000')
   

   if ((recgrid%z%n .eq. 2) .and. (recgrid%z%d .eq. 0)) call erexit('Expect over/under but found dz=0')
   
   if (bc%wfmout) then
      if ((nnz-nw)*vel%x%n*vel%y%n*nt/jsnap*4 .gt. 1.e10) call erexit('size of wfm is too big, pls increase jsnap')
   end if

   !! Definition
   vel%smooth = smooth
  
  !! hard-coded for now
   nhx=30
   dhx = vel%x%d
   ohx = -nhx*dhx
   
   nax = 60.
   dax = 0.06
   oax = -1.8
!   dax=4.
!   oax=-nax*dax

   sougrid%x%n = 1
   sougrid%y%n = 1
   nnz = vel%z%n
   nnx = vel%x%n
   nny = vel%y%n
   bndx = bc%bndx
   bndy = bc%bndy
   bndz = bc%bndz
   vel%z%n = vel%z%n + 2*bndz; 
   vel%x%n = vel%x%n+2*bndx
   vel%y%n = vel%y%n+2*bndy
   nz = vel%z%n;
   nx = vel%x%n;
   ny = vel%y%n;
   nt=recgrid%t%n
   nw = (vel%zw-vel%z%o)/vel%z%d + .5
   ! redefine nrt if greater than tmax
   tmax = (recgrid%nt-1)*recgrid%dt+recgrid%ot
   nnrt = (tmax-recgrid%t%o)/recgrid%t%d + .5
   write(0,*) 'nnrt, nrt', nnrt, recgrid%t%n
   recgrid%t%n = min(nnrt, recgrid%t%n)
  
   !*********************************
   write(0,*) '**************************'
   write(0,*) 'vel: nz, dz, oz ', nz, vel%z%d, vel%z%o
   write(0,*) 'vel: nx, dx, ox',  nx, vel%x%d, vel%x%o
   write(0,*) 'vel: ny, dy, oy',  ny, vel%y%d, vel%y%o
   write(0,*) 'vel: zw, nw',      vel%zw, nw
   write(0,*) 'sou: nx, dx, ox',  sougrid%x%n, sougrid%x%d, sougrid%x%o  
   write(0,*) 'sou: ny, dy, oy',  sougrid%y%n, sougrid%y%d, sougrid%y%o  
   write(0,*) 'sou: bfile, efile', sougrid%bfile, sougrid%efile
   write(0,*) 'sou: ntsav, dtsav, otsav',     sougrid%ntsav, sougrid%dtsav, sougrid%otsav
   write(0,*) 'rec: nx, dx, ox',  recgrid%x%n, recgrid%x%d, recgrid%x%o
   write(0,*) 'rec: ny, dy, oy',  recgrid%y%n, recgrid%y%d, recgrid%y%o
   write(0,*) 'rec: nrt, drt, ort', recgrid%t%n, recgrid%t%d, recgrid%t%o
   write(0,*) 'mod: nt, dt, ot', recgrid%nt, recgrid%dt, recgrid%ot
   write(0,*) 'nhx, nrz',        nhx, recgrid%z%n
   write(0,*) 'nop, verb',        bc%nop,verb
   write(0,*) 'smooth, wfmout, wfrout', smooth, bc%wfmout, bc%wfrout
   write(0,*) ' zr',              recgrid%z%o
   write(0,*) 'bndz, bndx, bndy', bc%bndz, bc%bndx, bc%bndy
   write(0,*) 'lpfil, VDer, gauss', bc%lpfil, bc%VDer, bc%gauss
   write(0,*) 'DW, bmask, emask', bc%DW, bc%bmaskX, bc%emaskX, bc%bmaskY, bc%emaskY
   write(0,*) '**************************'


   !----------------------------------------------------------------

   
  !! output files (offset domain CIG)
   call auxputch("n1",'i', nnz-nw, trim(out_));   call auxputch("label1",'s', 'z', trim(out_)); 
   call auxputch("d1",'f', vel%z%d, trim(out_));  call auxputch("o1",'f', vel%zw, trim(out_));
   call auxputch("n2",'i', nnx, trim(out_));      call auxputch("label2",'s', 'x', trim(out_)); 
   call auxputch("d2",'f', vel%x%d, trim(out_));  call auxputch("o2",'f', vel%x%o, trim(out_));
   call auxputch("n3",'i', nny, trim(out_));      call auxputch("label3",'s', 'y', trim(out_)); 
   call auxputch("d3",'f', vel%y%d, trim(out_));  call auxputch("o3",'f', vel%y%o, trim(out_));
   call auxputch("n4",'i', 2*nhx+1, trim(out_));      call auxputch("label4",'s', 'hx', trim(out_)); 
   call auxputch("d4",'f', dhx, trim(out_));  call auxputch("o4",'f', ohx, trim(out_));

   !----------------------------------------------------------------
   ! Allocation
   write(0,*) 'allocating', recgrid%x%n, recgrid%y%n, nt
   allocate(data(recgrid%x%n,recgrid%y%n,nt))
   write(0,*) 'allocating size', nnx,nny,nnz-nw,nnx*nny*(nnz-nw)
   allocate(model((nnz-nw),nnx,nny,1+2*nhx))


   !----------------------------------------------------------------
   ! initialize
   do it = 1,nt
     do iy=1,recgrid%y%n
       do ix=1,recgrid%x%n
         data(ix,iy,it) = 0.;
       end do
     end do
   end do
   do iy = 1,nny
     do ix=1,nnx
       do iz=1,nnz-nw
         model(iz,ix,iy,:) = 0.;
       end do
     end do
   end do



   write(0,*) 'before reading rec data'
   ! Read in the rec data
   call sep_read(data, rec_)



   if (bc%wfmout) then

         call auxputch("n1",'i',nnz-nw,trim(wfldmod_));    call auxputch("label1", 's', 'Z', trim(wfldmod_));
         call auxputch("d1",'f',vel%z%d,trim(wfldmod_));   call auxputch("o1",'f',vel%z%o,trim(wfldmod_));

         call auxputch("n2",'i',nnx,trim(wfldmod_));   	 call auxputch("label2", 's', 'X', trim(wfldmod_));
         call auxputch("d2",'f',vel%x%d,trim(wfldmod_));   call auxputch("o2",'f',vel%x%o,trim(wfldmod_));

         call auxputch("n3",'i',nny,trim(wfldmod_));   	 call auxputch("label3", 's', 'Y', trim(wfldmod_));
         call auxputch("d3",'f',vel%y%d,trim(wfldmod_));   call auxputch("o3",'f',vel%y%o,trim(wfldmod_));

         call auxputch("n4",'i',nt/jsnap,trim(wfldmod_));   
         call auxputch("d4",'f',-1.*recgrid%t%d*jsnap,trim(wfldmod_));   
         call auxputch("o4",'f',recgrid%t%o,trim(wfldmod_));
 	 call auxputch("label4", 's', 'time', trim(wfldmod_));    

   end if


   
   ! Call L_initiate
     sstat=L_init(sougrid, recgrid, vel, nhx, wfldsou_,vp_,rho, bc, wfldrec_in=wfldrec_, & 
     wfldmod_in=wfldmod_, verb_in=verb,jsnap_in=jsnap, MWfile_in=MWfile_, DWfile_in=DWfile_,nodes_in=nodes)


   ! ***************************   
   !  Main Body of the program
   ! ***************************

      ! Adjoint only
      ! RTM: data ---> reflectivity 
      sstat=L_oper(.true.,.false.,model,data) 

    

      ! Output result
	write(0,*) 'Output model'
      call sep_write(model,out_,esize=4)     ! ODCIG


      call sep_close()
     
     deallocate(data, model);

end if

   ! Stop timer
   call itime(now2)
   write(0,*) 'compute time Mig_Main3D_Off (s)', 3600*(now2(1)-now1(1))+60.*(now2(2)-now1(2))+(now2(3)-now1(3))
  
 end program
 
