program MissingTrcMask
! Create out Masking file with 0's at dead trace and 1's
! Input: 3D data file (x,y,t)
! Output: 3D mask file (x,y,t)
! Find out where dead trace is in (x,y) plane by looking at the energy along the trace


  use sep
!  use waveprop3d_mod

  implicit none
  
  integer                   :: naxis
  integer                   :: n1,n2,n3
  integer                   :: i, i1, i2, i3
  real                      :: o1, o2, o3
  real                      :: d1, d2, d3
  real*8                    :: val, thres
  character(len=128)        :: label1, label2, label3
  character(len=128)        :: getfile_
  character(len=128)        :: file1_, outfile_
  real, dimension(:,:,:), allocatable   ::  Array, outArr

  call sep_init()

  call from_param("naxis",naxis);

  
  if ((naxis .gt. 5) .or. (naxis .lt. 1) ) call erexit('naxis must be between 1 to 5')
  
 
    
  call from_param("file1", file1_);
  call from_param("outfile", outfile_);

  getfile_ = file1_
  
  n1=1.; n2=1.; n3=1.; 


     call from_aux(getfile_, "n1", n1);    call from_aux(getfile_,"d1",d1);     call from_aux(getfile_, "o1",o1);
     call from_aux(getfile_, "label1",label1);
     call from_aux(getfile_, "n2", n2);    call from_aux(getfile_,"d2",d2);     call from_aux(getfile_, "o2",o2);
     call from_aux(getfile_, "label2",label2);
     call from_aux(getfile_, "n3", n3);    call from_aux(getfile_,"d3",d3,1.);     call from_aux(getfile_, "o3",o3,0.);
     call from_aux(getfile_, "label3",label3);
  

    allocate( Array(n1,n2,n3), outArr(n1,n2,n3))

  do i3=1,n3
     do i2=1,n2
        do i1=1,n1
           Array(i1,i2,i3) = 0.
           outArr(i1,i2,i3) = 0.
       end do
    end do
  end do

  call sep_read(Array, file1_);
   
  write(0,*) 'n1, n2, n3', n1, n2, n3



  ! Begin Dot product
     do i3=1,n3
        do i2=1,n2
           do i1=1,n1
              val = Array(i1,i2,i3) 
              outArr(i1,i2,1) = outArr(i1,i2,1) + val**2
           end do
        end do
     end do


  thres = 1. 

  do i2=1,n2
     do i1=1,n1
        val = outArr(i1,i2,1) 
        if (val .gt. thres) then
           outArr(i1,i2,:) = 1.
        else
           outArr(i1,i2,:) = 0.
        end if
     end do
  end do


  write(0,*) 'pass 2'

  call sep_write(outArr, trim(outfile_), esize=4);

     call auxputch( "n1", 'i', n1, trim(outfile_));    call auxputch("d1",'f',d1, trim(outfile_));     call auxputch( "o1",'f',o1, trim(outfile_));
     call auxputch( "label1",'s',label1, trim(outfile_));
     call auxputch( "n2", 'i', n2, trim(outfile_));    call auxputch("d2",'f',d2, trim(outfile_));     call auxputch( "o2",'f',o2, trim(outfile_));
     call auxputch( "label2",'s',label2, trim(outfile_));
     call auxputch( "n3", 'i', n3, trim(outfile_));    call auxputch("d3",'f',d3, trim(outfile_));     call auxputch( "o3",'f',o3, trim(outfile_));
     call auxputch( "label3",'s',label3, trim(outfile_));


  deallocate(Array, outArr)
end program
