program ParabolicTauPfilter
  ! Filter ADCIG in the Parabolic Tau-p domain 
  ! Basically try to get rid of all the curvy event
  ! Input: angle-domain CIG (before filter)
  !      : axis 1=z, 2=x 3=y, 4=ax
  ! Output: Angle-domain CIG (after filter)
  !      : axis 1=z, 2=x 3=y, 4=ax 
  ! Output: Angle-domain CIG (after filter)
  !      : axis 1=z, 2=x 3=y, 4=ax 

  use sep
  use omp_lib
  use parataup_mod
  use ifport

  implicit none

  logical(4) :: success      
  logical                   :: verb, taupout
  integer                   :: it, nt, n4, npzero
  integer                   :: nx, ny, nz, nax, nkz, np, nkp 
  integer                   :: ix, iy, iz, iax, ikz, ip, ikp 
  real                      :: dx, dy, dz, dax, dkz, dp, dkp 
  real                      :: ox, oy, oz, oax, okz, op, okp 
  real :: pi=4.*atan(1.)
  real, dimension(:,:), allocatable  :: realArr1, realArr2
  complex, dimension(:,:), allocatable  :: comArr1, comArr2, elseArr2, elseArr1, comArr3, comXtalk1
  real, dimension(:,:,:,:), allocatable  :: model, model2, elsemodel2, model3
  complex, dimension(:), allocatable     :: in1,in2,out1,out2
  complex, dimension(:,:), allocatable   :: cdata1, cdata2, sigma
  complex, dimension(:,:,:,:), allocatable  :: xtalkArr

  character(len=128)        :: out_, out2_, trans_, xtalk_

  ! OMP stuff 
  integer :: inode,node,nodes,Totalnodes


  call sep_init()


  !!======================================================
  !! Parameter reading
  
  write(0,*) ' This is the beginning of myOff2Ang'

   call from_param("verb",verb,.false.);      


  !! input file (angle domain CIG)
   call from_history("n1", nz);  
   call from_history("d1", dz);  call from_history("o1", oz);
   call from_history("n2", nx);      
   call from_history("d2", dx);  call from_history("o2", ox);
   call from_history("n3", ny);     
   call from_history("d3", dy);  call from_history("o3", oy);
   call from_history("n4", nax);
   call from_history("d4", dax);      call from_history("o4", oax);



   !! Auxillary output files
   call from_param("outtag", out_);
   call from_param("elseouttag", out2_);
   call from_param("taupout",taupout,.false.);
   if (taupout) call from_param("taupfile", trans_);

   ! Parallel Programming
   call from_param("nodes" ,nodes, 8)
   !nodes=8
   Totalnodes =omp_get_num_procs()
   call omp_set_num_threads(nodes); 
   write(0,*) 'OMP: using', nodes , 'out of', Totalnodes
    

  !! Set temporary binary datapath
!  success = setenvqq("DATAPATH=/tmp/")


    np = 161
    dp =0.015
    op=-1.2
    npzero = (0.-op)/dp
 
    nkz = nz
    okz = -1.*pi/(dz)
    dkz = 2.*pi/(nz*dz)

    nkp = np
    okp = -1.*pi/(dp)
    dkp = 2.*pi/(np*dp)

   !*********************************
   write(0,*) '**************************'
   write(0,*) ' nz, dz, oz ', nz, dz, oz
   write(0,*) ' nx, dx, ox',  nx, dx, ox
   write(0,*) ' ny, dy, oy',  ny, dy, oy
   write(0,*) ' nax, dax, oax',  nax, dax, oax
   write(0,*) ' np, dp, op',  np, dp, op
   write(0,*) ' nkz, dkz, okz', nkz, dkz, okz
   write(0,*) ' nkp, dkp, okp', nkp, dkp, okp
   write(0,*) '**************************'

   !----------------------------------------------------------------

  !! output file (angle domain CIG)
   call auxputch("n1",'i', nz, trim(out_));      call auxputch("label1",'s', 'z', trim(out_)); 
   call auxputch("d1",'f', dz, trim(out_));      call auxputch("o1",'f', oz, trim(out_));
   call auxputch("n2",'i', nx, trim(out_));      call auxputch("label2",'s', 'x', trim(out_)); 
   call auxputch("d2",'f', dx, trim(out_));      call auxputch("o2",'f', ox, trim(out_));
   call auxputch("n3",'i', ny, trim(out_));      call auxputch("label3",'s', 'y', trim(out_)); 
   call auxputch("d3",'f', dy, trim(out_));       call auxputch("o3",'f', oy, trim(out_));
   call auxputch("n4",'i', nax, trim(out_));  call auxputch("label4",'s', 'ax', trim(out_)); 
   call auxputch("d4",'f', dax, trim(out_));      call auxputch("o4",'f', oax, trim(out_));



   call auxputch("n1",'i', nz, trim(out2_));      call auxputch("label1",'s', 'z', trim(out2_)); 
   call auxputch("d1",'f', dz, trim(out2_));      call auxputch("o1",'f', oz, trim(out2_));
   call auxputch("n2",'i', nx, trim(out2_));      call auxputch("label2",'s', 'x', trim(out2_)); 
   call auxputch("d2",'f', dx, trim(out2_));      call auxputch("o2",'f', ox, trim(out2_));
   call auxputch("n3",'i', ny, trim(out2_));      call auxputch("label3",'s', 'y', trim(out2_)); 
   call auxputch("d3",'f', dy, trim(out2_));       call auxputch("o3",'f', oy, trim(out2_));
   call auxputch("n4",'i', nax, trim(out2_));  call auxputch("label4",'s', 'ax', trim(out2_)); 
   call auxputch("d4",'f', dax, trim(out2_));      call auxputch("o4",'f', oax, trim(out2_));

  if (taupout) then
  !! output tau-p transformed file
   call auxputch("n1",'i', nz, trim(trans_));      call auxputch("label1",'s', 'z', trim(trans_)); 
   call auxputch("d1",'f', dz, trim(trans_));      call auxputch("o1",'f', oz, trim(trans_));
   call auxputch("n2",'i', nx, trim(trans_));      call auxputch("label2",'s', 'x', trim(trans_)); 
   call auxputch("d2",'f', dx, trim(trans_));      call auxputch("o2",'f', ox, trim(trans_));
   call auxputch("n3",'i', ny, trim(trans_));      call auxputch("label3",'s', 'y', trim(trans_)); 
   call auxputch("d3",'f', dy, trim(trans_));       call auxputch("o3",'f', oy, trim(trans_));
   call auxputch("n4",'i', np, trim(trans_));  call auxputch("label4",'s', 'p', trim(trans_)); 
   call auxputch("d4",'f', dp, trim(trans_));      call auxputch("o4",'f', op, trim(trans_));
   end if

   !----------------------------------------------------------------
   ! Allocation

   write(0,*) 'allocating size', nx,ny,nz,nax,nx*ny*nz*(nax)
   allocate(model(nz,nx,ny,nax))
   write(0,*) 'pass 1'
   allocate(model2(nz,nx,ny,nax))
   allocate(elsemodel2(nz,nx,ny,nax))
   write(0,*) 'pass 2'
   allocate(model3(nz,nx,ny,np))
   write(0,*) 'pass 3'
   allocate(realArr1(nz,nax),realArr2(nz,np))
   write(0,*) 'pass 4'
   allocate(comArr1(nkz,nax), comArr2(nkz,np), elseArr1(nkz,nax), elseArr2(nkz,np), comArr3(nkz,nkp))
   write(0,*) 'pass 5'
   allocate(cdata1(nz,nax),cdata2(nkz,np),sigma(nkz,np))
   allocate(in1(nkz),out1(nkz),in2(np),out2(np))   
!----------------------------------------------------------------
   ! initialize
   do iy = 1,ny
     do ix=1,nx
       do iz=1,nz
         model(iz,ix,iy,:) = 0.;
         model2(iz,ix,iy,:) = 0.;
         elsemodel2(iz,ix,iy,:) = 0.;
         model3(iz,ix,iy,:) = 0.;
       end do
     end do
   end do

   write(0,*) 'pass 6'

   ! Read in ODCIG
   call sep_read(model)
   write(0,*) 'pass 7'

   ! initialization


   call ft_init(nkz, nax,in1,out1,in2,out2) 
   call ft_init2(nkz, np,in1,out1,in2,out2) 
   call sigma_init(sigma,nax,dax,oax,nkz,dkz,okz,np,dp,op,in2,out2)

write(0,*) 'set exclude negative kp'

	do iy=1,ny
!	do iy=ny/4,3*ny/4

         write(0,*) 'calculationg iy=', iy, 'out of', ny

	!$OMP PARALLEL DO DEFAULT(shared) schedule(dynamic,1) num_threads(nodes) PRIVATE(inode, ix,iax,iz,realArr1,comArr1,comArr2,comArr3,realArr2,cdata1,cdata2,in1,in2,out1,out2) 
        do ix=1,nx

  	     inode = omp_get_thread_num()



!         write(0,*) 'calculationg ix=', ix, 'out of', nx, 'node', inode
           do iax = 1,nax
              do iz=1,nz   
                 realArr1(iz,iax) = model(iz,ix,iy,iax)
!                 realArr1(iz,iax) = model(iz,ix,iy,iax)*xtalkArr(iz,ix,iy,iax)
!		 comXtalk1(iz,:) = xtalkArr(iz,ix,iy,:)
              end do
           end do



        ! Forward FFT in z dir: from U(z,ax) to U(kz,ax)         
        call forwardFFTinZ(realArr1,comArr1,cdata1,nz,oz,dz,nax,in1,out1) 


        ! Apply Equation (26) to do forward tau-p transform
        ! Vp'(kz,p) = sum_x U(kz,ax) e^(i kz p ax^2)
	call forwardPara_oper(comArr1,comArr2,nkz,dkz,okz,np,dp,op,nax,dax,oax)


        ! Forward FFT in p direction: from Vp'(kz,p) to Vp'(kz,kp)
!	call forwardFFTinP(comArr2,comArr3,cdata2,np,op,dp,nkz,in2,out2)

        ! Deconvolution Vp(kz,kp) = Vp'(kz,kp) 1/(2pi) sqrt(kzkp) Vp'(kz,kp)
!	call paraDecon(comArr3,nkz,dkz,okz,np,dp,op)

	! inverse FFT in p direction: from Vp(kz,kp) back to Vp(kz,p)
!        call inverseFFTinP(comArr3,comArr2,cdata2,np,op,dp,nkz,in2,out2)

	call AngleDomain_filter(np,npzero,op,dp,nz,oz,dz,comArr2,elseArr2)


	call inversePara_oper(comArr2,comArr1,nkz,dkz,okz,np,dp,op,nax,dax,oax)
        call inversePara_oper(elseArr2,elseArr1,nkz,dkz,okz,np,dp,op,nax,dax,oax)

	! output the post-filtered array
        call ft1axis(.true.,1.,comArr1,in1,out1)       
        call ft1axis(.true.,1.,elseArr1,in1,out1)       

	! output the tau-p array
        call ft1axis2(.true.,1.,comArr2,in1,out1)       
        do ip=1,np
           do iz=1,nz
               realArr2(iz,ip) = real(comArr2(iz,ip))
           end do
        end do 


             !$OMP CRITICAL
           do iax = 1,nax
              do iz=1,nz   
                 model2(iz,ix,iy,iax) = real(comArr1(iz,iax))
                 elsemodel2(iz,ix,iy,iax) = real(elseArr1(iz,iax))
              end do
           end do


           do ip = 1,np
              do iz=1,nz   
                 model3(iz,ix,iy,ip) = realArr2(iz,ip)
              end do
           end do
             !$OMP END CRITICAL

        end do
	!$OMP END PARALLEL DO
      end do
                      

      ! Output result
	write(0,*) 'Output model'
      call sep_write(model2,out_,esize=4)   ! ADCIG
      call sep_write(elsemodel2,out2_,esize=4)   ! ADCIG
      if (taupout) then
         call sep_write(model3,trans_,esize=4)   
      end if

      call sep_close()
     
     deallocate(model,model2,model3);



 end program
 
