program RemoveTop

! Output a file where the top eps portion is zerored
! Input: infile.H
! Input: eps (tolerance)
! Output: outfile.H
! If sub=true, output = file1 less the same values

  use sep

implicit none




 logical    :: verb, sub
 integer    :: naxis,i, nvolume
 integer    :: locval, ntop, itop
 integer    :: i1, i2, i3, i4, i5
 integer    :: n1, n2, n3, n4, n5
 real       :: o1, o2, o3, o4, o5
 real       :: d1, d2, d3, d4, d5
 real       :: eps, val1, val2, thres
 character(len=128)        :: label1, label2, label3, label4, label5
 character(len=128)        :: infile_,  outfile_
  real, dimension(:), allocatable   :: inArray, tmpArray, outArray
  call sep_init()


  call from_param("naxis",naxis);
  call from_param("eps",eps);
  call from_param("sub",sub,.false.);
  
  if ((naxis .gt. 5) .or. (naxis .lt. 1) ) call erexit('naxis must be between 1 to 5')
  
  call from_param('outfile',outfile_) 
  call from_param("infile", infile_);



  
  n1=1.; n2=1.; n3=1.;  n4=1.; n5=1.;

  if(naxis .ge. 1) then
     call from_aux(infile_, "n1", n1);    call from_aux(infile_,"d1",d1);     call from_aux(infile_, "o1",o1);
     call from_aux(infile_, "label1",label1);
  end if
  
  if (naxis .ge. 2) then
     call from_aux(infile_, "n2", n2);    call from_aux(infile_,"d2",d2);     call from_aux(infile_, "o2",o2);
     call from_aux(infile_, "label2",label2);
  end if
  
  if(naxis .ge. 3) then
     call from_aux(infile_, "n3", n3);    call from_aux(infile_,"d3",d3,1.);     call from_aux(infile_, "o3",o3,0.);
     call from_aux(infile_, "label3",label3);
  end if
  
  if (naxis .ge. 4) then
     call from_aux(infile_, "n4", n4);    call from_aux(infile_,"d4",d4);     call from_aux(infile_, "o4",o4);
     call from_aux(infile_, "label4",label4);
  end if
  
  if (naxis .ge. 5) then
     call from_aux(infile_, "n5", n5);    call from_aux(infile_,"d5",d5);     call from_aux(infile_, "o5",o5);
     call from_aux(infile_, "label5",label5);
  end if
  
  nvolume=n1*n2*n3*n4*n5
  allocate(outArray(nvolume), inArray(nvolume), tmpArray(nvolume))
  do i=1,nvolume
    outArray(i) = 0.;
    inArray(i) = 0.;
    tmpArray(i) = 0.;
  end do

  call sep_read(inArray, infile_);

  ntop = eps*nvolume
  do i=1,nvolume
     tmpArray(i) = abs(inArray(i))
  end do
  
  do itop = 1,ntop   
    if (mod(itop,2000).eq.0) write(0,*) 'itop=',itop,' out of', ntop
    locval =  maxloc(tmpArray,1)
    tmpArray(locval) = 0.
    outArray(locval) = inArray(locval)

  end do  

  if (sub) then
    write(0,*) 'subtract'
    do i=1,nvolume
       val1 = outArray(i); 
       outArray(i) = inArray(i) - val1
    end do
  end if

    if (naxis .ge. 1) then
         call auxputch("n1",'i',n1, trim(outfile_));     
         call auxputch("d1",'f',d1, trim(outfile_));   
         call auxputch("o1",'f',o1,trim(outfile_));
         call auxputch("label1",'s',label1,trim(outfile_));
     end if
     if (naxis .ge. 2) then
	  call auxputch("n2",'i',n2, trim(outfile_));     
	  call auxputch("d2",'f',d2, trim(outfile_));   
	  call auxputch("o2",'f',o2,trim(outfile_));
          call auxputch("label2",'s',label2,trim(outfile_));
     end if
     if (naxis .ge. 3) then
	  call auxputch("n3",'i',n3, trim(outfile_));     
	  call auxputch("d3",'f',d3, trim(outfile_));   
	  call auxputch("o3",'f',o3,trim(outfile_));
          call auxputch("label3",'s',label3,trim(outfile_));
     end if
     if (naxis .ge. 4) then
	  call auxputch("n4",'i',n4, trim(outfile_));     
	  call auxputch("d4",'f',d4, trim(outfile_));   
	  call auxputch("o4",'f',o4,trim(outfile_));
          call auxputch("label4",'s',label4,trim(outfile_));
     end if
     if (naxis .ge. 5) then
	  call auxputch("n5",'i',n5, trim(outfile_));     
	  call auxputch("d5",'f',d5, trim(outfile_));   
	  call auxputch("o5",'f',o5,trim(outfile_));
          call auxputch("label5",'s',label5,trim(outfile_));
     end if
     
     call sep_write(outArray, outfile_);


end program
