program SecondDer3D

! Take Second derivative of input file
! It has 3 options
! VDer =  [0] vertical derivative only (d2/dz2)
! lpfil = [0] take laplacian 
! gauss = [0] take gaussian version of laplacian
! only one of 'VDer', 'lpfil', and 'gauss' should be on.
! if not, 'gauss' take precedence over 'lpfil', 
! and 'lpfil' take precendence over 'VDer'

  use sep
  use waveprop3d_mod
  use type3d_mod

implicit none


 type(velo) :: mm

 logical    :: adj, VDer, lpfil, gauss
 integer    :: nx, nz, ny,n4
 real       :: d4,o4
 real, dimension(:), allocatable  ::  temp
 real, dimension(:,:,:,:), allocatable :: model

 call sep_init()


   write(0,*) ' This is SecondDer3D program'


   ! read in type of filter
   call from_param("adj",adj,.true.);
   call from_param("lpfil",lpfil, .false.);
   call from_param("VDer",VDer, .false.);         
   call from_param("gauss",gauss, .false.);



  ! Read back ground mmocity param
   call from_history("n1", mm%z%n);  call from_history("d1", mm%z%d);  call from_history("o1",mm%z%o);
   call from_history("n2", mm%x%n);  call from_history("d2", mm%x%d);  call from_history("o2",mm%x%o);
   call from_history("n3", mm%y%n);  call from_history("d3", mm%y%d);  call from_history("o3",mm%y%o);
     call from_history("n4", n4,1);  call from_history("d4", d4,1.);  call from_history("o4",o4,0.);


   nz = mm%z%n;
   nx = mm%x%n;
   ny = mm%y%n;

   write(0,*) 'nz,dz,oz', mm%z%n, mm%z%d, mm%z%o
   write(0,*) 'nx,dx,ox', mm%x%n, mm%x%d, mm%x%o
   write(0,*) 'ny,dy,oy', mm%y%n, mm%y%d, mm%y%o
   write(0,*) 'gauss, lpfil, VDer', gauss, lpfil, VDer
   write(0,*) 'adj', adj


   allocate( model(nz,nx,ny,n4))


write(0,*) 'pass 1'

   ! Read in the model
    call sep_read( model)

    if (gauss) then
	 call Gaussian3DExtended(adj,model,mm%z%n,mm%x%n,mm%y%n,n4)
    else if (lpfil) then
	 call  Laplacian3DExtended(adj,model,mm%z%n,mm%x%n,mm%y%n,n4)
    else if (VDer) then
        call Ver2ndDer3DExtended(adj,model,mm%z%n,mm%x%n,mm%y%n,n4)
    end if

write(0,*) 'pass 3'
   
!----------------------------------------------
! Output

  call sep_write(model,dim1=nz, dim2=nx, dim3=ny, dim4=n4, esize=4)

write(0,*) 'pass 4'
  
  call to_history("n1",mm%z%n); call to_history("d1",mm%z%d);  call to_history("o1",mm%z%o);
  call to_history("n2",mm%x%n); call to_history("d2",mm%x%d);  call to_history("o2",mm%x%o);
  call to_history("n3",mm%y%n); call to_history("d3",mm%y%d);  call to_history("o3",mm%y%o);
  call to_history("n4",n4); call to_history("d4",d4);  call to_history("o4",o4);  

write(0,*) 'pass 5'

end program
