program SouGen3Ddn
! Pre-compute source wavefield with specification defined in a parameter file
! Source wavefield is saved in  file 
! Reflecting top boundary
! And sharp velocity jump
!
! zw    : depth of the waterbody
! zsb   : depth at the bottom of the seabed
! 
!


  use sep
  use omp_lib
  use waveprop3d_mod
  use type3d_mod
  use smooth3d_mod
  use ifport

  implicit none
  
  
  type(sou_grid)       :: sougrid
  type(velo)           :: vel_grid
  type(bc_values)      :: bc


  character(len=128)   :: wfldsou_,sou_, vfile_
    

  logical(4) :: success    
  logical                              :: verb, wfldout, smooth, ASout, abctop
  integer                              :: isave, nsave, xsnap, mode, ifile, ntmp
  integer                              :: nx, ix, nnx, ixs, ixsloc, bndx
  integer                              :: ny, iy, nny, iys, iysloc, bndy
  integer                              :: nz, iz, nnz, izs, izsloc, bndz
  integer                              :: bndztop1, bndzbot1, bndztop2, bndzbot2
  integer                              :: nop, jsnap, ndata, nskipcom, nskipsav
  integer                              :: nst, nsb, ndelay,  nt, nw, it, iw
  integer                              :: iot, ioot, isnap, nlast
  real                                 :: dx, ox, dsx, osx, idx, Gx
  real                                 :: dy, oy, dsy, osy, idy, Gy
  real                                 :: dz, oz, dsz, osz, idz, Gz
  real                                 :: dt, dt2, ot, tdelay
  real                                 :: amp, rho, watvel, mxv, minvel, usize1
  real                                 :: xtapco, ytapco, ztapco
  real, dimension(:), allocatable      :: xtaper, xtaper2, tempArr
  real, dimension(:), allocatable      :: ytaper, ytaper2
  real, dimension(:), allocatable      :: ztaper, ztaper2
  real, dimension(:), allocatable      :: fs, wvl
  real, dimension(:,:,:), allocatable  :: vp2, Vo, ampArr,  ArealSou
  real, dimension(:,:,:), allocatable   :: uo1, um1, up1
  real, dimension(:,:,:,:),allocatable :: usave
  character(len=128)                   :: out_, swout_, ampfile_, ASfile_

  ! OMP stuff 
  integer :: inode,node,nodes

  ! Expanding Domain Stuff
  integer :: nzmax
  real    :: t, vmax

  ! Blocky stuff
  integer ::  zbsz, xbsz, ybsz

  ! Timer stuff 
  integer*4   :: time1(3), time2(3) 
  call itime(time1)

  call sep_init()

!!======================================================
  !! Parameter reading

   write(0,*) 'read 1'

   call from_param("verb",verb,.false.);   
   call from_param("wfldout", wfldout,.false.);   call from_param("ASout", ASout,.false.);
   call from_param("abctop",abctop,.true.);       call from_param("smooth", smooth, .false.);
   call from_param("mode", mode, 1); 


   write(0,*) 'read 2'

   ! Read source and out file
   call from_param("soufile", sou_);   
   call from_param("outtag", out_);
   call from_param("ampfile", ampfile_);
   call from_param("asfile", asfile_);
   call from_param("wfldsou", wfldsou_, "wfldsou.H");
   call from_param("vfile", vfile_);


   write(0,*) 'read 3', vfile_

   ! Read back ground velocity param
   call from_aux(vfile_, "n1", vel_grid%z%n);  call from_aux( vfile_, "d1", vel_grid%z%d);  
   call from_aux( vfile_, "o1",vel_grid%z%o);
   call from_aux( vfile_, "n2", vel_grid%x%n);  call from_aux( vfile_, "d2", vel_grid%x%d);  
   call from_aux( vfile_, "o2",vel_grid%x%o);
   call from_aux( vfile_, "n3", vel_grid%y%n);  call from_aux( vfile_, "d3", vel_grid%y%d);  
   call from_aux( vfile_, "o3",vel_grid%y%o);
  

   write(0,*) 'read 4'
 
   ! Read water depth, water vel and density
   call from_param('rho', rho);
   call from_param('zw', vel_grid%zw);               ! Depth of the water column
   call from_param('zsb', vel_grid%zsb);             ! depth of the bottom of the seabed zone
   

   write(0,*) 'read 5'

   ! Source arrangement
  
   ! Read position of the source from command line
   call from_param('Gx',Gx);     call from_param('Gy',Gy);     call from_param('Gz',Gz);
   izsloc =  (Gz-vel_grid%z%o)/vel_grid%z%d - 1.5
   ixsloc =  (Gx-vel_grid%x%o)/vel_grid%x%d + .5
   iysloc =  (Gy-vel_grid%y%o)/vel_grid%y%d + .5

   
   call from_aux(sou_, "n1", sougrid%t%n);  call from_aux(sou_,"d1", sougrid%t%d);  


   write(0,*) 'read 6'
   
   ! Read time information
   call from_param("nt",nt);  call from_param("dt",dt);   call from_param("ot",ot)
   call from_param("tdelay",tdelay);


   write(0,*) 'read 7'
   
   !! taper coefficients and padding
   call from_param("xtapco",bc%xtapco,0.0002);  
   call from_param("ztapco",bc%ztapco,0.0002)
   call from_param("ytapco",bc%ytapco,0.0002); 
   call from_param("nop", bc%nop,8);
   call from_param("bndz",bc%bndz,bc%nop); if(bc%bndz<bc%nop) bc%bndz=bc%nop; 
   call from_param("bndx",bc%bndx,bc%nop); if(bc%bndx<bc%nop) bc%bndx=bc%nop;
   call from_param("bndy",bc%bndy,bc%nop); if(bc%bndy<bc%nop) bc%bndy=bc%nop;


   write(0,*) 'read 8'

   !! Auxillary output files parameter
   call from_Param("jsnap", jsnap, 100);
   call from_param("xsnap", xsnap, 8);


   !! OMP Stuff
   call from_param("nodes" ,nodes, 8)
   node =omp_get_num_procs()
   call omp_set_num_threads(nodes); 
   write(0,*) 'OMP using', nodes, 'threads out of' , node


   write(0,*) 'read 9'

   !! Blocky stuff
   call from_param("Zblocksz",zbsz,40)
   call from_param("Xblocksz",xbsz,40)
   call from_param("Yblocksz",ybsz,40)

   !! Vmax stuff
   call from_param("vmax",vmax,4500.)


   write(0,*) 'read 10'

  !! Set temporary binary datapath
!   success = setenvqq("DATAPATH=/tmp/")

    nnx = vel_grid%x%n;  nny = vel_grid%y%n;  nnz = vel_grid%z%n;
    dx = vel_grid%x%d;   dy = vel_grid%y%d;   dz = vel_grid%z%d;
    ox = vel_grid%x%o;   oy = vel_grid%y%o;   oz = vel_grid%z%o;
    bndx = bc%bndx;        bndy = bc%bndy;      bndz = bc%bndz;
    xtapco = bc%xtapco;  ytapco = bc%ytapco;  ztapco = bc%ztapco
    nx = nnx + 2*bndx;   ny = nny + 2*bndy;  

    nop = bc%nop

    if (abctop)then
       bndztop2= bndz
       bndzbot2= bndz
    else
       bndztop2 = nop
       bndzbot2 = bndz
    end if

    nz = nnz + bndztop2 + bndzbot2

    nw = (vel_grid%zw - oz)/dz + .5
    nsb = (vel_grid%zsb - vel_grid%zw - oz)/dz + 0.5
    ndelay = (tdelay - ot)/dt + 1.5  
    nskipcom = (vel_grid%zw/1450.)/dt+0.5
    nskipsav =((Gz+vel_grid%zw)/1450.)/dt+0.5
    if (nskipcom .lt. ndelay) nskipcom = 0
    if (nskipsav .lt. ndelay) nskipsav = 0
    isnap = mod(1+nskipsav,xsnap)
    iot = max(ndelay +1 + mod(xsnap-mod(ndelay+1,xsnap)+1,xsnap),nskipsav+1)
    nsave = (nt+ndelay-iot)/xsnap+1.5; isave = 1
    nlast = ((nt+ndelay-iot)/xsnap)*xsnap+iot

 !*********************************
   write(0,*) '**************************'
   write(0,*) 'vel: nz, dz, oz ', nnz, nz, vel_grid%z%d, vel_grid%z%o
   write(0,*) 'vel: nx, dx, ox',  nnx, nx, vel_grid%x%d, vel_grid%x%o
   write(0,*) 'vel: ny, dy, oy',  nny, ny, vel_grid%y%d, vel_grid%y%o 
   write(0,*) 'vel: zw, nw',      vel_grid%zw, nw
   write(0,*) 'vel: zsb, nsb',    vel_grid%zsb, nsb
   write(0,*) 'Gx, Gy, Gz',           Gx, Gy, Gz
   write(0,*) 'tdelay, ndelay',  tdelay, ndelay
   write(0,*) 'nskipsav',        nskipsav
   write(0,*) 'nsave, nskipcom, nlast', nsave, nskipcom, nlast
   write(0,*) 'iot',                 iot
   write(0,*) 'ixsloc, iysloc, izsloc',      ixsloc, iysloc, izsloc
   write(0,*) 'nt, ot, dt',       nt, ot, dt
   write(0,*) 'nbc%nop, verb', bc%nop, verb
   write(0,*) ' wfldout',         wfldout
   write(0,*) 'bndz, bndx, bndy, nop',  bndz, bndx, bndy, nop
   write(0,*) 'ztapco, xtapco, ytapco',   ztapco, xtapco, ytapco
   write(0,*) '# of nodes used',  nodes
   write(0,*) 'abctop, smooth, ASout, mode',    abctop, smooth, ASout, mode
   write(0,*) 'blocky stuff ...........'
   write(0,*) 'vmax, bsz:z,x,y', vmax, zbsz, xbsz, ybsz
   write(0,*) '**************************'



   !! Error checking
   if ((mode .gt. 2) .or. (mode .lt. 1)) call erexit('Mode must be 1 or 2')
   if ( (nz-nw)*nnx*nny*nsave*4 .gt. 50.e9 ) call erexit('Sou wavefield is more than 50G')

 !******************************************************************
 ! Part Zero: Read in velocity model first

    !Allocate & initialization
 
    allocate(vp2(nz,nx,ny)) 

    do iy=1,ny
      do ix=1,nx
        do iz=1,nz      
          vp2(iz,ix,iy) = 0.    
        end do
      end do
    end do

   ! Read in the background velocity
    call sep_read( vp2(1+bndztop2:nz-bndzbot2,1+bndx:nx-bndx,1+bndy:ny-bndy), vfile_)

   ! Padding
    do iz=1,bndztop2; vp2(iz,:,:) = vp2(1+bndztop2,:,:); end do
    do iz=1,bndzbot2; vp2(nz-iz+1,:,:) = vp2(nz-bndzbot2,:,:); end do 
    do ix=1,bndx; vp2(:,ix,:) = vp2(:,1+bndx,:); vp2(:,nx-ix+1,:) = vp2(:,nx-bndx,:); end do 
    do iy=1,bndy; vp2(:,:,iy) = vp2(:,:,1+bndy); vp2(:,:,ny-iy+1) = vp2(:,:,ny-bndy); end do 

    do iy=1,ny
      do ix=1,nx
        do iz=1,nz      
          vp2(iz,ix,iy) = vp2(iz,ix,iy)*vp2(iz,ix,iy)      
        end do
      end do
    end do



   nst = nt 
   allocate(fs(nst), wvl(nst))
   fs = 0.;

   ! Read in the source wavelet
   call sep_read( fs(1:sougrid%t%n), sou_)

   do iw=1+1,sougrid%t%n-1;  wvl(iw) = (-2.*fs(iw) + fs(iw+1) + fs(iw-1)); end do

   mxv = max(maxval(wvl),abs(minval(wvl)))
   wvl = wvl/mxv
   wvl(sougrid%t%n:nst)=0.;
       
!  call sep_write(fs, "fs.H");
!  call sep_write(wvl, "fs.H");
!   call auxputch("n1",'i',nst,'fs.H');  
!   call auxputch("d1",'f',0.002,'fs.H');   
!   call auxputch("o1",'f',-0.150,'fs.H');
!   call auxputch("n2",'i',2,'fs.H');
!   call auxputch("d2",'f',1.,'fs.H');
!   call auxputch("o2",'f',0.,'fs.H');
!  call erexit('Exit here for debug')


  !******************************************************************

! This is for part I

    bndztop1=nop; bndzbot1=bndz;
    nz = nw + nsb + bndztop1 + bndzbot1             ! pad the top



    write(0,*) 'nz, nw, nsb', nz, nw, nsb
    write(0,*) 'bndz, bndx bndy', nop, bndz, bndx, bndy
    write(0,*) 'bndztop1, bndzbot1, bndz', bndztop1, bndzbot1

   ! Memory Allocation   
   allocate(ztaper(bndzbot1))
   allocate(xtaper(bndx), ytaper(bndy))
   allocate(ArealSou(nnx,nny,nst))
   

   allocate(up1(nz,nx,ny), um1(nz,nx,ny), uo1(nz,nx,ny))
   allocate(Vo(nz,nx,ny))


 
   do iz=bndztop1+1,nz-bndz
      Vo(iz,:,:) = vp2(iz,:,:)
   end do
   
   do iz=1,bndzbot1; Vo(nz-iz+1,:,:) = Vo(nz-bndzbot1,:,:); end do !padding
   do iz=1,bndztop1; Vo(iz,:,:) = Vo(bndztop1+1,:,:); end do !padding     
   

    amp=0.05
    
    !! Absorbing BC and Tapersetup  

    call  tapersetupCos3d(bndzbot1,bndx,bndy,ztaper,xtaper,ytaper,0.98,0.02)
    
    dt2 = dt*dt
    idx = dt2/dx**2; idy=dt2/dy**2; idz = dt2/dz**2;
 
    write(0,*) 'pass 1'
 

   call auxputch("n1",'i',nnz-nw, trim(ampfile_)); 
   call auxputch("d1",'f', dz, trim(ampfile_));  call auxputch("o1",'f', oz, trim(ampfile_));
   call auxputch("n2",'i', nnx, trim(ampfile_)); 
   call auxputch("d2",'f', dx, trim(ampfile_));  call auxputch("o2",'f', ox, trim(ampfile_));
   call auxputch("n3",'i', nny, trim(ampfile_)); 
   call auxputch("d3",'f', dy, trim(ampfile_));  call auxputch("o3",'f', oy, trim(ampfile_));
   call auxputch("n4",'i', 1, trim(ampfile_)); 
   call auxputch("d4",'f', 1., trim(ampfile_));  
   call auxputch("o4",'f', 0., trim(ampfile_));               
   call auxputch("label1", 's', 'nz', trim(ampfile_));
   call auxputch("label2", 's', 'nx', trim(ampfile_));
   call auxputch("label3", 's', 'ny', trim(ampfile_));
   call auxputch("label4", 's', 'UD', trim(ampfile_));

    write(0,*) 'pass 2'

 if (ASout .and. (mode .ne. 2) ) then

   call auxputch("n1",'i',nnx, trim(ASfile_)); 
   call auxputch("d1",'f', dx, trim(ASfile_));  call auxputch("o1",'f', ox, trim(ASfile_));
   call auxputch("n2",'i', nny, trim(ASfile_)); 
   call auxputch("d2",'f', dy, trim(ASfile_));  call auxputch("o2",'f', oy, trim(ASfile_));
   call auxputch("n3",'i', nt, trim(ASfile_)); 
   call auxputch("d3",'f', dt, trim(ASfile_));  call auxputch("o3",'f', -tdelay, trim(ASfile_));             
   call auxputch("label1", 's', 'nx', trim(ASfile_));
   call auxputch("label2", 's', 'ny', trim(ASfile_));
   call auxputch("label3", 's', 'nst', trim(ASfile_));


 end if

    write(0,*) 'pass 3'

 if (wfldout) then
 	   call auxputch("n1",'i',nnz+bndztop2+bndz,trim(wfldsou_));
   	   call auxputch("d1",'f',dz,trim(wfldsou_));   
	   call auxputch("o1",'f',vel_grid%zw-(bndz)*dz, trim(wfldsou_));
           call auxputch("n2",'i',nx,trim(wfldsou_));  
   	   call auxputch("d2",'f',dx,trim(wfldsou_));   
	   call auxputch("o2",'f',vel_grid%x%o-(bndx)*dx, trim(wfldsou_));
           call auxputch("n3",'i',ny,trim(wfldsou_));  
   	   call auxputch("d3",'f',dy,trim(wfldsou_));   
	   call auxputch("o3",'f',vel_grid%y%o-(bndy)*dy, trim(wfldsou_));
   	   call auxputch("label1", 's', 'Z', trim(wfldsou_));
     	   call auxputch("label2", 's', 'X', trim(wfldsou_));
     	   call auxputch("label3", 's', 'Y', trim(wfldsou_));
! temporary
   call auxputch("n4",'i', (nt+ndelay-iot)/jsnap+1.5, trim(wfldsou_)); 
   call auxputch("d4",'f', dt*xsnap, trim(wfldsou_));   
   call auxputch("o4",'f', ((iot-1)*dt-tdelay), trim(wfldsou_));  
           call auxputch("o4",'f', -tdelay+(jsnap-1)*dt, trim(wfldsou_));            
  	   call auxputch("label4", 's', 'time', trim(wfldsou_));

 end if

    write(0,*) 'pass 4'

    
       ixs = bndx + ixsloc
       iys = bndy + iysloc
       izs = izsloc + bndztop1 + 0

       ! initializing the array
       do iy=1,ny
         do ix=1,nx
           do iz=1,nz
             up1(iz,ix,iy)=0.; um1(iz,ix,iy)=0.; uo1(iz,ix,iy)=0.;
           end do          
         end do
       end do

    write(0,*) 'pass 5'

      if (mode .eq. 1) then

       do it=1,nst

        !  write(0,*) '1-it', it

          !! 16th order laplacian
          call laplac3d16(.true., uo1,up1, Vo, nx, ny,nz, nz, xbsz,ybsz,zbsz, idx,idy,idz, nop,it,nodes)

        !  write(0,*) '2-it', it

          !! time step
          do iy=1,ny
            do ix=1,nx
               do iz=1,nz
                 up1(iz,ix,iy)=up1(iz,ix,iy)-um1(iz,ix,iy)
                 um1(iz,ix,iy)=uo1(iz,ix,iy); 
                 uo1(iz,ix,iy)=up1(iz,ix,iy);
               end do          
            end do
          end do
 
        !  write(0,*) '3-it', it, izs, ixs, iys

          !! Injecting source/rec wavelet into the finite diff grid    
          uo1(izs,ixs,iys) =  uo1(izs,ixs,iys) - wvl(it)  


        !  write(0,*) '4-it', it
           
          !! taper
	  call taperingCos3dV2(nz,nx,ny,bndztop1,bndzbot1,bndx,bndy,ztaper,xtaper,ytaper, uo1)
	  call taperingCos3dV2(nz,nx,ny,bndztop1,bndzbot1,bndx,bndy,ztaper,xtaper,ytaper, up1)
	  call taperingCos3dV2(nz,nx,ny,bndztop1,bndzbot1,bndx,bndy,ztaper,xtaper,ytaper, um1)


          !! Store areal source to memory for every  time step
          if (it .lt. nt-1) then
            do iy=1,nny
              do ix=1,nnx
                 ArealSou(ix,iy,it) = uo1(bndztop1+1,bndx+ix,bndy+iy) 
	      end do
            end do	
          end if     

        !  write(0,*) '5-it', it

          if (verb .and. (mod(it,50) .eq. 1)) then
             write(0,*) 'Pre-Sou: it, sum', it, sum(uo1), sum(ArealSou)
          end if
        
      
!         if (mod(it,jsnap)==0 .and. wfldout ) then
!           if (it .gt. 50 .and. wfldout) then
!             write(0,*) 'uo1, dim', size(uo1,1), size(uo1,2), it
!             call sep_write(uo1,wfldsou_)
!           end if
!          end if
        

       end do !(end source time loop)
       
       ! Taperoff the edges of ArealSou (in x,y, and t)  to avoid sharp cutoff      
       allocate(ztaper2(bndzbot1))
       allocate(xtaper2(20), ytaper2(20))
       call  tapersetupCos3d(bndz,20,20,ztaper2,xtaper2,ytaper2,0.5,0.5)
       call  taperingCos3d(nnx,nny,nt,20,20,bndz,xtaper2,ytaper2,ztaper2, ArealSou)    

       deallocate(xtaper2,ytaper2,ztaper2)

       if (ASout) then
	  write(0,*) ' Writing AS file' 
       	  call sep_write(ArealSou , trim(ASfile_), dim1=nnx, dim2=nny, dim3=nst, esize=4)
       end if

  
    end if

    deallocate(uo1,um1,up1)

    
    
 !******************************************************************
 !******************************************************************
    ! This is part 2 

    write(0,*) 'pass 6' 

    if (mode .eq. 2) then    
       call sep_read(ArealSou(1:nnx,1:nny,1:nst), trim(ASfile_))
    end if
 
    ifile=1
    nz = nnz + bndztop2 + bndzbot2             ! pad the top
    izs = bndztop2+1


    write(0,*) 'allocating 3', nz*nx*ny
    allocate(up1(nz,nx,ny), um1(nz,nx,ny), uo1(nz,nx,ny))


 ! The program will try to allocating no more than 10 Gz to store SW
    ! for furthur time steps, it will be written to disk on several run
    usize1 = 1.0*(nnz-nw)
    usize1 = usize1*nnx
    usize1 = usize1*nny 
    ntmp = 2.5e8/usize1-1. 
    ntmp = min(ntmp,nsave)
    write(0,*) 'allocating usave', (nnz-nw), nnx, nny, ntmp
    write(0,*) 'allocating usave', usize1*ntmp, usize1*ntmp*4./1.0e9,'GB'
    allocate(usave(nnz-nw,nnx,nny,ntmp))
    write(0,*) 'size of usave', size(usave,1), size(usave,2), size(usave,3), size(usave,4)

    write(0,*) 'allocating', (nz-nw)*nx*ny
    allocate(ampArr(nz-nw,nx,ny))

    ! initializing the array
    do iy=1,ny
      do ix=1,nx
           up1(1:nz,ix,iy) = 0.
           um1(1:nz,ix,iy) = 0.
           up1(1:nz,ix,iy) = 0.
           ampArr(1:nz-nw,ix,iy) = 0.       
      end do
    end do
       
    do it=1+nskipcom,nlast
  
      !! Expanding grid calculation
      t = (it-1)*dt+ot
!     nzmax = int(max(zbsz,min(((vmax*t-oz)/dz + bndztop2+3*nop + 1),real(nz))))
      nzmax = nz


      !! 16th order laplacian
      call laplac3d16(.true., uo1,up1, vp2, nx, ny, nz, nzmax, xbsz,ybsz,zbsz, idx,idy,idz, nop,it,nodes)


      !! time step
      do iy=1,ny
        do ix=1,nx
          do iz=1,nz
             up1(iz,ix,iy)=up1(iz,ix,iy)-um1(iz,ix,iy)
             um1(iz,ix,iy)=uo1(iz,ix,iy); 
             uo1(iz,ix,iy)=up1(iz,ix,iy);
          end do
        end do     
      end do




      !! Injecting source/rec wavelet into the finite diff grid
      if (it .lt. nt) then 
        do iy=1,nny
          do ix=1,nnx 
             uo1(izs,ix+bndx,iy+bndy) = uo1(izs,ix+bndx,iy+bndy) + -1*amp*ArealSou(ix,iy,it)
       	  end do
        end do
      end if


      !! Tapering 
      if (abctop) then
  	 call taperingCos3d(nz,nx,ny,bndz,bndx,bndy,ztaper,xtaper,ytaper,uo1)
	 call taperingCos3d(nz,nx,ny,bndz,bndx,bndy,ztaper,xtaper,ytaper,up1)
	 call taperingCos3d(nz,nx,ny,bndz,bndx,bndy,ztaper,xtaper,ytaper,um1)
      else
	 call taperingCos3dV2(nz,nx,ny,bndztop2,bndzbot2,bndx,bndy,ztaper,xtaper,ytaper, uo1)
	 call taperingCos3dV2(nz,nx,ny,bndztop2,bndzbot2,bndx,bndy,ztaper,xtaper,ytaper, up1)
	 call taperingCos3dV2(nz,nx,ny,bndztop2,bndzbot2,bndx,bndy,ztaper,xtaper,ytaper, um1)
      end if

 
      !! Saving the amplitude file (for preconditioning/constraint later)
      do iy=1,nny
	do ix=1,nnx
          do iz=1,nnz-nw
             ampArr(iz,ix,iy) = ampArr(iz,ix,iy) + uo1(iz+nop+nw,ix+bndx,iy+bndy)**2
          end do
        end do
      end do


     ! Save the sou-side wavefield at every xsnap time step
     if ((it .gt. nskipsav) .and. (it .gt. ndelay).and.((mod(it-ndelay,xsnap)==0) .or. (xsnap .eq. 1)) ) then
         if (isave .eq. 1) ioot =it
         write(0,*) 'saving to usave', it, it-ndelay, isave, sum(uo1)

         do iy=1,nny
           do ix=1,nnx
             do iz=1,nnz-nw
                usave(iz,ix,iy,isave) = uo1(bndztop2+nw+iz,bndx+ix,bndy+iy)
             end do
           end do
         end do
	 isave = isave+1            
     end if

      if (verb .and. (mod(it,50) .eq. 1)) write(0,*) 'Sou: it, sum', it-ndelay, sum(uo1)
 
    ! Write SouWfld to disk and then reset isave to 1
     if (isave .ge. (ntmp+1)) then
 
         write(0,*) 'write to disk', it, ifile, nsave
         call getfilename(out_,swout_,ifile)

         call auxputch("n1",'i',nnz-nw, trim(swout_)); 
         call auxputch("d1",'f', dz, trim(swout_));  
         call auxputch("o1",'f', vel_grid%zw, trim(swout_));
         call auxputch("n2",'i', nnx, trim(swout_));  
         call auxputch("d2",'f', dx, trim(swout_));  
         call auxputch("o2",'f', ox, trim(swout_));
         call auxputch("n3",'i', nny, trim(swout_));   
         call auxputch("d3",'f', dy, trim(swout_));  
         call auxputch("o3",'f', oy, trim(swout_));
         call auxputch("n4",'i', ntmp, trim(swout_)); 
         call auxputch("d4",'f', dt*xsnap, trim(swout_));   
         call auxputch("o4",'f', ((ioot-1)*dt-tdelay), trim(swout_));     
         call auxputch("label1", 's', 'nz', trim(swout_));
         call auxputch("label2", 's', 'nx', trim(swout_));
         call auxputch("label3", 's', 'ny', trim(swout_));
         call auxputch("label4", 's', 'nt', trim(swout_));

	 call sep_write(usave, trim(swout_), dim1=nnz-nw, dim2=nnx, dim3=nny, dim4=ntmp,esize=4)
         ! reset isave 
 !         if (ifile .ge. 20) call erexit('Exit after first write')
         isave = 1
         ifile = ifile+1
      end if
           

   end do !(end source time loop)


! ------------------------------------------------------------
! ------ Output file -----------------------------------------

     call sep_write(ampArr, trim(ampfile_), dim1=nnz-nw,dim2=nnx,dim3=nny,esize=4)
 
!-----------------------------------------------------

   deallocate( fs, wvl)
   write(0,*) 'dealloc 1'
   deallocate(vp2)
   write(0,*) 'dealloc 2'
   deallocate( ztaper, ytaper, xtaper)
   write(0,*) 'dealloc 3'
   deallocate( up1, um1, uo1)
   write(0,*) 'dealloc 4'
   deallocate( ArealSou)
   write(0,*) 'dealloc 5'
   deallocate( usave)
   call sep_close()

! Stop timer
  call itime(time2)
    write(0,*) 'compute time SouGen3Ddn (s)', 3600*(time2(1)-time1(1))+60.*(time2(2)-time1(2))+(time2(3)-time1(3))


write(0,*) 'program should exit'

end program
