program SouGen3Dup
! Pre-compute source wavefield with specification defined in a parameter file
! Source wavefield is saved in the out_ file 
! This is for Up-going signal 
! And sharp velocity jump
!
! zw    : depth of the waterbody
! zsb   : depth at the bottom of the seabed
! 
!


  use sep
  use omp_lib
  use septimers
  use waveprop3d_mod
  use type3d_mod
  use ifport

  implicit none
  
  
  type(sou_grid)       :: sougrid
  type(velo)           :: vel_grid
  type(bc_values)      :: bc


  character(len=128)   :: wfldsou_,sou_, vfile_
    
  logical(4) :: success        
  logical                              :: verb, wfldout
  integer                              :: nsave, ntmp, xsnap, isave, ifile
  integer                              :: nx, ix, nnx, ixs, ixsloc, bndx
  integer                              :: ny, iy, nny, iys, iysloc, bndy
  integer                              :: nz, iz, nnz, izs, izsloc, bndz
  integer                              :: bndztop, bndzbot
  integer                              :: nop, jsnap, ndata, iot
  integer                              :: nst, nsb, ndelay,  nt, nw, it, iw
  real                                 :: dx, ox, dsx, osx, idx, Gx
  real                                 :: dy, oy, dsy, osy, idy, Gy
  real                                 :: dz, oz, dsz, osz, idz, Gz
  real                                 :: dt, dt2, ot, tdelay
  real                                 :: amp, rho, watvel, mxv, val
  real                                 :: xtapco, ytapco, ztapco
  real*8                               :: usize1, usize2
  real, dimension(:), allocatable      :: xtaper, ytaper, ztaper
  real, dimension(:), allocatable      :: fs, wvl
  real, dimension(:,:,:), allocatable  :: vp2, Vo, ampArr
  real, dimension(:,:,:), allocatable  :: uo1, um1, up1
  real, dimension(:,:,:,:), allocatable  :: usave
  character(len=128)                   :: out_, swout_, ampfile_

  ! OMP stuff 
  integer :: inode,node,nodes

  ! Expanding Domain Stuff
  integer :: nzmax
  real    :: t, vmax

  ! Wavefield truncation + tapering
  integer :: ntsave
  real    :: vmin, zmax, ymax, xmax, Dmax

  ! Blocky stuff
  integer ::  zbsz, xbsz, ybsz

  ! Initial timer
  integer*4                 :: now2(3), now1(3)
  call itime(now1)

  call sep_init()

!!======================================================
  !! Parameter reading

   call from_param("verb",verb,.false.);   
   call from_param("wfldout", wfldout,.false.);   


   ! Read source and aux file name
   call from_param("soufile", sou_);   
   call from_param("outtag", out_);
   call from_param("ampfile", ampfile_);
   call from_param("wfldsou", wfldsou_, "wfldsou.H");
   call from_param("vfile", vfile_);
   
   ! Read back ground velocity param
   call from_aux(vfile_, "n1", vel_grid%z%n);  call from_aux( vfile_, "d1", vel_grid%z%d);  
   call from_aux( vfile_, "o1",vel_grid%z%o);
   call from_aux( vfile_, "n2", vel_grid%x%n);  call from_aux( vfile_, "d2", vel_grid%x%d);  
   call from_aux( vfile_, "o2",vel_grid%x%o);
   call from_aux( vfile_, "n3", vel_grid%y%n);  call from_aux( vfile_, "d3", vel_grid%y%d);  
   call from_aux( vfile_, "o3",vel_grid%y%o);



   ! Read water depth, water vel and density
   call from_param('rho', rho);
   call from_param('zw', vel_grid%zw);               ! Depth of the water column
   call from_param('zsb', vel_grid%zsb);             ! depth of the bottom of the seabed zone
   

   ! Read position of the source from command line
   call from_param('Gx',Gx);     call from_param('Gy',Gy);     call from_param('Gz',Gz);
   izsloc =  (Gz-vel_grid%z%o)/vel_grid%z%d + .5
   ixsloc =  (Gx-vel_grid%x%o)/vel_grid%x%d + .5
   iysloc =  (Gy-vel_grid%y%o)/vel_grid%y%d + .5

   
   call from_aux(sou_, "n1", sougrid%t%n);  call from_aux(sou_,"d1", sougrid%t%d);  


   
   ! Read time information
   call from_param("nt",nt);  call from_param("dt",dt);   call from_param("ot",ot)
   call from_param("tdelay",tdelay,0.);

   
   !! taper coefficients and padding
   call from_param("xtapco",bc%xtapco,0.0002);  
   call from_param("ztapco",bc%ztapco,0.0002)
   call from_param("ytapco",bc%ytapco,0.0002); 
   call from_param("nop", bc%nop,8);
   call from_param("bndz",bc%bndz,bc%nop); if(bc%bndz<bc%nop) bc%bndz=bc%nop; 
   call from_param("bndx",bc%bndx,bc%nop); if(bc%bndx<bc%nop) bc%bndx=bc%nop;
   call from_param("bndy",bc%bndy,bc%nop); if(bc%bndy<bc%nop) bc%bndy=bc%nop;

   !! Auxillary output files parameter
   call from_param("jsnap", jsnap, 100);
   call from_param("xsnap", xsnap, 8);


   !! OMP Stuff
   call from_param("nodes" ,nodes, 4)
   node =omp_get_num_procs()
   call omp_set_num_threads(nodes); 
   write(0,*) 'OMP using', nodes, 'threads out of' , node

   !! Blocky stuff
   call from_param("Zblocksz",zbsz,40)
   call from_param("Xblocksz",xbsz,40)
   call from_param("Yblocksz",ybsz,40)

   !! Vmax stuff
   call from_param("vmax",vmax,4500.); call from_param("vmin",vmin,1480.)

  !! Set temporary binary datapath
  !success = setenvqq("DATAPATH=/tmp/")

    nnx = vel_grid%x%n;  nny = vel_grid%y%n;  nnz = vel_grid%z%n;
    dx = vel_grid%x%d;   dy = vel_grid%y%d;   dz = vel_grid%z%d;
    ox = vel_grid%x%o;   oy = vel_grid%y%o;   oz = vel_grid%z%o;
    bndx = bc%bndx;        bndy = bc%bndy;      bndz = bc%bndz;
    xtapco = bc%xtapco;  ytapco = bc%ytapco;  ztapco = bc%ztapco
    nx = nnx + 2*bndx;   ny = nny + 2*bndy;  

    nop = bc%nop
    bndztop= bndz
    bndzbot= bndz

    nw = (vel_grid%zw - oz)/dz + .5
    nsb = (vel_grid%zsb - vel_grid%zw - oz)/dz + 0.5
    ndelay = (tdelay - ot)/dt + 1.5
    xmax = vel_grid%x%o + (nnx-1)*vel_grid%x%d
    ymax = vel_grid%y%o + (nny-1)*vel_grid%y%d
    zmax = vel_grid%z%o + (nnz-1)*vel_grid%z%d
    Dmax = max(abs(Gx-vel_grid%x%o),abs(xmax-Gx),abs(zmax-Gz),abs(Gy-vel_grid%y%o),abs(ymax-Gy))
    ntsave = ((Dmax/vmin)/dt)+1.5
    nsave = ntsave/xsnap; isave=1 
    ifile = 1; iot=1

    write(0,*) 'xmax, ymax, zmax', xmax, ymax, zmax
    write(0,*) 'Dmax, vmin, ntsave', Dmax, vmin, ntsave
 


 !*********************************
   write(0,*) '**************************'
   write(0,*) 'vel: nz, dz, oz ', nnz, nz, vel_grid%z%d, vel_grid%z%o
   write(0,*) 'vel: nx, dx, ox',  nnx, nx, vel_grid%x%d, vel_grid%x%o
   write(0,*) 'vel: ny, dy, oy',  nny, ny, vel_grid%y%d, vel_grid%y%o 
   write(0,*) 'vel: zw, nw',      vel_grid%zw, nw
   write(0,*) 'vel: zsb, nsb',    vel_grid%zsb, nsb
   write(0,*) 'Gx, Gy, Gz',           Gx, Gy, Gz
   write(0,*) 'tdelay, ndelay',           tdelay, ndelay
   write(0,*) 'jsnap, xsnap, ntsave, nsave',  jsnap, xsnap, ntsave, nsave
   write(0,*) 'ixsloc, iysloc, izsloc',      ixsloc, iysloc, izsloc
   write(0,*) 'nt, ot, dt',       nt, ot, dt
   write(0,*) 'nbc%nop, verb', bc%nop, verb
   write(0,*) ' wfldout',         wfldout
   write(0,*) 'bndz, bndx, bndy, nop',  bndz, bndx, bndy, nop
   write(0,*) 'ztapco, xtapco, ytapco',   ztapco, xtapco, ytapco
   write(0,*) '# of nodes used',  nodes
   write(0,*) 'blocky stuff ...........'
   write(0,*) 'vmax, bsz:z,x,y', vmax, zbsz, xbsz, ybsz
   write(0,*) '**************************'


   !! Error checking


 !******************************************************************
 ! Part Zero: Read in velocity model first

  write(0,*) 'pass 1'

    !Allocate & initialization
 
    allocate(vp2(nnz,nnx,nny)) 

  write(0,*) 'pass 2'

    do iy=1,nny
      do ix=1,nnx
        do iz=1,nnz      
          vp2(iz,ix,iy) = 0.    
        end do
      end do
    end do

  write(0,*) 'pass 3'

   ! Read in the background velocity
    call sep_read( vp2(1:nnz,1:nnx,1:nny),vfile_)

  write(0,*) 'pass 4'

  !$OMP PARALLEL DO private(iy,ix,iz,val) schedule(dynamic,1) num_threads(nodes)
    do iy=1,nny
      do ix=1,nnx
        do iz=1,nnz 
          val = vp2(iz,ix,iy)     
          vp2(iz,ix,iy) = val**2      
        end do
      end do
    end do
  !$OMP END PARALLEL DO

  write(0,*) 'pass 5'

   nst = nt 
   allocate(fs(nst),wvl(nst))
   fs = 0.; wvl=0.;

   ! Read in the source wavelet
   call sep_read( fs(1:sougrid%t%n), sou_)

   do iw=1+1,sougrid%t%n-1;  wvl(iw) = (-2.*fs(iw) + fs(iw+1) + fs(iw-1)); end do

   mxv = max(maxval(wvl),abs(minval(wvl)))
   wvl = wvl/mxv
   wvl(sougrid%t%n:nst)=0.;


  !******************************************************************


    nz = nnz - nw + bndztop + bndzbot             ! pad the top


    write(0,*) 'nz, nw, nnz', nz, nw, nnz
    write(0,*) 'bndz, bndx bndy', nop, bndz, bndx, bndy
    write(0,*) 'bndztop, bndzbot, bndz', bndztop, bndzbot

   ! Memory Allocation   
   allocate(ztaper(bndz))
   allocate(xtaper(bndx), ytaper(bndy))
  

   allocate(Vo(nz,nx,ny))
 
write(0,*) 'bndx, bndy', bndx, bndy
 
   do iy=1,nny
     do ix=1,nnx
       do iz=1,nnz-nw
      Vo(iz+bndztop,ix+bndx,iy+bndz) = vp2(iz+nw,ix,iy)
       end do
     end do
   end do
   
   ! Padding
    do iz=1,bndztop; Vo(iz,:,:) = Vo(1+bndztop,:,:); end do
    do iz=1,bndzbot; Vo(nz-iz+1,:,:) = Vo(nz-bndzbot,:,:); end do 
    do ix=1,bndx; Vo(:,ix,:) = Vo(:,1+bndx,:); Vo(:,nx-ix+1,:) = Vo(:,nx-bndx,:); end do 
    do iy=1,bndy; Vo(:,:,iy) = Vo(:,:,1+bndy); Vo(:,:,ny-iy+1) = Vo(:,:,ny-bndy); end do   


    amp=0.05
    
    !! Absorbing BC and Tapersetup  

    call  tapersetupCos3d(bndzbot,bndx,bndy,ztaper,xtaper,ytaper,0.98,0.02)
    
    dt2 = dt*dt
    idx = dt2/dx**2; idy=dt2/dy**2; idz = dt2/dz**2;
 
    write(0,*) 'pass 1'


   call auxputch("n1",'i',nnz-nw, trim(ampfile_)); 
   call auxputch("d1",'f', dz, trim(ampfile_));  call auxputch("o1",'f', oz, trim(ampfile_));
   call auxputch("n2",'i', nnx, trim(ampfile_)); 
   call auxputch("d2",'f', dx, trim(ampfile_));  call auxputch("o2",'f', ox, trim(ampfile_));
   call auxputch("n3",'i', nny, trim(ampfile_)); 
   call auxputch("d3",'f', dy, trim(ampfile_));  call auxputch("o3",'f', oy, trim(ampfile_));
   call auxputch("n4",'i', 1, trim(ampfile_)); 
   call auxputch("d4",'f', 1., trim(ampfile_));  
   call auxputch("o4",'f', 0., trim(ampfile_));               
   call auxputch("label1", 's', 'nz', trim(ampfile_));
   call auxputch("label2", 's', 'nx', trim(ampfile_));
   call auxputch("label3", 's', 'ny', trim(ampfile_));
   call auxputch("label4", 's', 'UD', trim(ampfile_));


 if (wfldout) then
 	   call auxputch("n1",'i',nz,trim(wfldsou_));
   	   call auxputch("d1",'f',dz,trim(wfldsou_));   
	   call auxputch("o1",'f',vel_grid%zw-(bndz)*dz, trim(wfldsou_));
           call auxputch("n2",'i',nx,trim(wfldsou_));  
   	   call auxputch("d2",'f',dx,trim(wfldsou_));   
	   call auxputch("o2",'f',vel_grid%x%o-(bndx)*dx, trim(wfldsou_));
           call auxputch("n3",'i',ny,trim(wfldsou_));  
   	   call auxputch("d3",'f',dy,trim(wfldsou_));   
	   call auxputch("o3",'f',vel_grid%y%o-(bndy)*dy, trim(wfldsou_));
   	   call auxputch("label1", 's', 'Z', trim(wfldsou_));
     	   call auxputch("label2", 's', 'X', trim(wfldsou_));
     	   call auxputch("label3", 's', 'Y', trim(wfldsou_));
           call auxputch("n4",'i',(nt+jsnap)/jsnap,trim(wfldsou_));
           call auxputch("d4",'f', dt*jsnap, trim(wfldsou_));   
           call auxputch("o4",'f', -tdelay+(jsnap-1)*dt, trim(wfldsou_));            
  	   call auxputch("label4", 's', 'time', trim(wfldsou_));

 end if


    ixs = bndx + ixsloc + 1
    iys = bndy + iysloc + 1
!    izs = izsloc -nw + bndztop 
    izs = izsloc -nw + bndztop + 5

    write(0,*) 'ixs, iys, izs', ixs, iys, izs
    write(0,*) 'nz, nx, ny, bndztop', nz, nx, ny, bndztop

    write(0,*) 'allocating 3', nz, nx, ny
    write(0,*) 'allocating 3', nz*nx*ny
    allocate(up1(nz,nx,ny), um1(nz,nx,ny), uo1(nz,nx,ny))

    ! The program will try to allocating no more than 10 Gz to store SW
    ! for furthur time steps, it will be written to disk on several run
    usize1 = 1.0*(nnz-nw)
    usize1 = usize1*nnx
    usize1 = usize1*nny 
    ntmp = 2.5e8/usize1-1. 
    ntmp = min(ntmp,nsave)
    write(0,*) 'allocating usave', (nnz-nw), nnx, nny, ntmp
    write(0,*) 'allocating usave', usize1*ntmp, usize1*ntmp*4./1.0e9,'GB'
    allocate(usave(nnz-nw,nnx,nny,ntmp))
    write(0,*) 'size of usave', size(usave,1), size(usave,2), size(usave,3), size(usave,4)
    allocate(ampArr(nz-nw,nx,ny))

write(0,*) 'pass 1'

    ! initializing the array
    do iy=1,ny
      do ix=1,nx
           uo1(1:nz,ix,iy) = 0.
           um1(1:nz,ix,iy) = 0.
           up1(1:nz,ix,iy) = 0.
           ampArr(1:nz-nw,ix,iy) = 0.       
      end do
    end do

write(0,*) 'pass 2'

    do it=1,ntmp
    do iy=1,nny
      do ix=1,nnx
         do iz=1,(nnz-nw)
           usave(iz,ix,iy,it) = 0.
         end do      
      end do
    end do
    end do

write(0,*) 'pass 3'
    ! -------------------------------------------------------------
    ! Pre-compute source wavefield
       
    do it=1,ntsave+ndelay
  
      !! Expanding grid calculation
      t = (it-1)*dt+ot
   !   nzmax = int(max(zbsz,min(((vmax*t-oz)/dz + izs+3*nop + 1),real(nz))))
      nzmax = nz		

! write(0,*) 'it ', it,  nzmax

      !! 16th order laplacian
      call laplac3d16(.true., uo1,up1, Vo, nx, ny, nz, nzmax, xbsz,ybsz,zbsz, idx,idy,idz, nop,it,nodes)

! write(0,*) '@it ', it, sum(uo1), sum(up1), sum(um1)

      !! time step
      do iy=1,ny
        do ix=1,nx
          do iz=1,nz
             up1(iz,ix,iy)=up1(iz,ix,iy)-um1(iz,ix,iy)
             um1(iz,ix,iy)=uo1(iz,ix,iy); 
             uo1(iz,ix,iy)=up1(iz,ix,iy);
          end do
        end do     
      end do

! write(0,*) '@@it ', it, sum(uo1), sum(up1), sum(um1), wvl(it)

      !! Injecting source/rec wavelet into the finite diff grid
      if (it .lt. 2*ndelay) then
         uo1(izs,ixs,iys) =  uo1(izs,ixs,iys) - wvl(it)  
      end if         

! write(0,*) '@@@it ', it, sum(uo1), sum(up1), sum(um1)

      !! Tapering 
      call taperingCos3d(nz,nx,ny,bndz,bndx,bndy,ztaper,xtaper,ytaper,uo1)
      call taperingCos3d(nz,nx,ny,bndz,bndx,bndy,ztaper,xtaper,ytaper,up1)
      call taperingCos3d(nz,nx,ny,bndz,bndx,bndy,ztaper,xtaper,ytaper,um1)

! write(0,*) '@@@@it ', it, sum(uo1), sum(up1), sum(um1)

      !! Calculate source side amplitude for preconditioning
      do iy=1,nny
	do ix=1,nnx
          do iz=1,nnz-nw
             ampArr(iz,ix,iy) = ampArr(iz,ix,iy) + uo1(iz+nop+nw,ix+bndx,iy+bndy)**2
          end do
        end do
      end do


     ! Save the sou-side wavefield at every xsnap time step
      if ((it .gt. ndelay).and.((mod(it-ndelay,xsnap)==0) .or. (xsnap .eq. 1)) ) then
         write(0,*) 'saving to usave', it, it-ndelay, isave, sum(uo1)
         if (isave .eq. 1) iot =it
!!	!$OMP PARALLEL DO default(shared) private(inode,iy,ix,iz) schedule(dynamic,1) num_threads(nodes)
         do iy=1,nny
           do ix=1,nnx
             do iz=1,nnz-nw
                usave(iz,ix,iy,isave) = uo1(bndztop+iz,bndx+ix,bndy+iy)
             end do
           end do
         end do
!! 	!$OMP END PARALLEL DO
	 isave = isave+1
      end if


      if (verb .and. (mod(it,50) .eq. 1)) write(0,*) 'Sou: it, sum', it-ndelay, sum(uo1)

      ! Write SouWfld to disk and then reset isave to 1
      if (isave .ge. (ntmp+1)) then
 
         write(0,*) 'write to disk', it, ifile, nsave
         call getfilename(out_,swout_,ifile)

         call auxputch("n1",'i',nnz-nw, trim(swout_)); 
         call auxputch("d1",'f', dz, trim(swout_));  
         call auxputch("o1",'f', vel_grid%zw, trim(swout_));
         call auxputch("n2",'i', nnx, trim(swout_));  
         call auxputch("d2",'f', dx, trim(swout_));  
         call auxputch("o2",'f', ox, trim(swout_));
         call auxputch("n3",'i', nny, trim(swout_));   
         call auxputch("d3",'f', dy, trim(swout_));  
         call auxputch("o3",'f', oy, trim(swout_));
         call auxputch("n4",'i', ntmp, trim(swout_)); 
         call auxputch("d4",'f', dt*xsnap, trim(swout_));   
         call auxputch("o4",'f', ((iot-1)*dt-tdelay), trim(swout_));     
         call auxputch("label1", 's', 'nz', trim(swout_));
         call auxputch("label2", 's', 'nx', trim(swout_));
         call auxputch("label3", 's', 'ny', trim(swout_));
         call auxputch("label4", 's', 'nt', trim(swout_));

	 call sep_write(usave, trim(swout_), dim1=nnz-nw, dim2=nnx, dim3=nny, dim4=ntmp,esize=4)
         ! reset isave 
!          if (ifile .ge. 20) call erexit('Exit after first write')
         isave = 1
         ifile = ifile+1
      end if


      if (mod(it,jsnap)==1 .and. wfldout) then
         write(0,*) 'writing to wfldsou', it, ndelay
	 call sep_write(uo1,wfldsou_, dim1=nz, dim2=nx,dim3=ny,esize=4)
      end if

   end do !(end source time loop)


! ------------------------------------------------------------
! ------ Output file -----------------------------------------
     call sep_write(ampArr, trim(ampfile_), dim1=nnz-nw,dim2=nnx,dim3=nny,esize=4)
 
!-----------------------------------------------------

   deallocate( fs, wvl)
   write(0,*) 'dealloc 1'
   deallocate(vp2)
   write(0,*) 'dealloc 2'
   deallocate( ztaper, ytaper, xtaper)
   write(0,*) 'dealloc 3'
   deallocate( up1, um1, uo1)
   write(0,*) 'dealloc 4'
   deallocate( usave)

   
   call sep_close()

! Stop timer
  call itime(now2)
  write(0,*) 'compute time SouGen3Dup (s)', 3600*(now2(1)-now1(1))+60.*(now2(2)-now1(2))+(now2(3)-now1(3))

write(0,*) 'program should exit'

end program
