program VDerfilter
  ! Forward FFT in z
  ! inv=false   apply (kz)^2 
  ! inv=true  apply (kz)^(-2) + eps
  ! Inverse FFT in z

  ! Input: 4D image cube (before filter)
  !      : axis 1=z, 2=x 3=y, 4=ax
  ! Output: 4D image cube (after filter)
  !      : axis 1=z, 2=x 3=y, 4=ax 

  use sep
  use omp_lib
  use fourier_mod
  use ifport

  implicit none

  logical(4) :: success      
  logical                   :: verb, inv
  integer                   :: it, nt, n4
  integer                   :: nx, ny, nz, nax, nkz
  integer                   :: ix, iy, iz, iax, ikz
  real                      :: dx, dy, dz, dax, dkz
  real                      :: ox, oy, oz, oax, okz 
  real                      :: pi=4.*atan(1.), eps, kz, sign1
  real, dimension(:,:), allocatable  :: realArr1, realArr2
  complex, dimension(:,:), allocatable  :: comArr1, comArr2
  real, dimension(:,:,:,:), allocatable  :: model, model2
  character(len=128)        :: out_

  ! OMP stuff 
  integer :: inode,node,nodes,Totalnodes


  call sep_init()


  !!======================================================
  !! Parameter reading
  
  write(0,*) ' This is the beginning of VDerfilter'

   call from_param("verb",verb,.false.);      
   call from_param("inv",inv,.false.);
   call from_param("eps",eps, 0.00001);

  !! input file (angle domain CIG)
   call from_history("n1", nz);  
   call from_history("d1", dz);  call from_history("o1", oz);
   call from_history("n2", nx);      
   call from_history("d2", dx);  call from_history("o2", ox);
   call from_history("n3", ny);     
   call from_history("d3", dy);  call from_history("o3", oy);
   call from_history("n4", nax,1);
   call from_history("d4", dax,1.);      call from_history("o4", oax,0.);






   ! Parallel Programming
   call from_param("nodes" ,nodes, 8)
   !nodes=8
   Totalnodes =omp_get_num_procs()
   call omp_set_num_threads(nodes); 
    write(0,*) 'OMP: using', nodes , 'out of', Totalnodes
    
 
    nkz = nz
    okz = -1.*pi/(dz)
    dkz = 2.*pi/(nz*dz)
    sign1 = 1.

   !*********************************
   write(0,*) '**************************'
   write(0,*) ' nz, dz, oz ', nz, dz, oz
   write(0,*) ' nx, dx, ox',  nx, dx, ox
   write(0,*) ' ny, dy, oy',  ny, dy, oy
   write(0,*) ' nax, dax, oax',  nax, dax, oax
   write(0,*) ' nkz, dkz, okz', nkz, dkz, okz
   write(0,*) ' eps, sign1', eps, sign1
   write(0,*) ' inv', inv
   write(0,*) '**************************'

   !----------------------------------------------------------------

  !! output file (angle domain CIG)
   call to_history("n1", nz);      call to_history("label1", 'z'); 
   call to_history("d1", dz);      call to_history("o1", oz);
   call to_history("n2", nx);      call to_history("label2", 'x'); 
   call to_history("d2", dx);      call to_history("o2", ox);
   call to_history("n3", ny);      call to_history("label3", 'y'); 
   call to_history("d3", dy);       call to_history("o3", oy);
   call to_history("n4", nax);  call to_history("label4", 'ax'); 
   call to_history("d4", dax);      call to_history("o4", oax);

   !----------------------------------------------------------------
   ! Allocation

   write(0,*) 'allocating size', nx,ny,nz,nax,nx*ny*nz*(nax)
   allocate(model(nz,nx,ny,nax))
   write(0,*) 'pass 1'
   allocate(model2(nz,nx,ny,nax))
   write(0,*) 'pass 2'
   allocate(realArr1(nz,nax))
   write(0,*) 'pass 3'
   allocate(comArr1(nkz,nax))
   write(0,*) 'pass 4'

   !----------------------------------------------------------------
   ! initialize
   do iy = 1,ny
     do ix=1,nx
       do iz=1,nz
         model(iz,ix,iy,:) = 0.;
         model2(iz,ix,iy,:) = 0.;
       end do
     end do
   end do

   write(0,*) 'pass 6'

   ! Read in ODCIG
   call sep_read(model)
   write(0,*) 'pass 7'

   ! initialization


   call ft_init(nkz, nax) 

	do iy=1,ny
!         write(0,*) 'calculationg iy=', iy, 'out of', ny

        do ix=1,nx
!         write(0,*) 'calculationg ix=', ix, 'out of', nx
           do iax = 1,nax
              do iz=1,nz   
                 comArr1(iz,iax) = cmplx(model(iz,ix,iy,iax),0.)
              end do
           end do

        ! Forward transform in z
	call ft1axis(.true.,sign1,comArr1)

        ! Apply kz filters
          do ikz=1,nkz
              kz=okz+(ikz-1)*dkz
              if (inv) then
                 comArr1(ikz,:) = comArr1(ikz,:)/(kz+eps)**2
              else
                 comArr1(ikz,:) = comArr1(ikz,:)*kz**2
	      end if
          end do !(ends kz loop)


        ! Inverse transform in z
	call ft1axis(.false.,sign1,comArr1)
           
           do iax = 1,nax
              do iz=1,nz   
                 model2(iz,ix,iy,iax) = real(comArr1(iz,iax))
              end do
           end do

        end do ! (ends x loop)
      end do ! (ends y loop)
                      

      ! Output result
	write(0,*) 'Output model'
      call sep_write(model2,esize=4)   ! ADCIG

      call sep_close()
     
     deallocate(model,model2);



 end program
 
