module fourier_mod2

  implicit none
  include "fftw3.f"

  complex, dimension(:), allocatable, private  :: iin1, oout1, iin2, oout2
  integer, private                             :: nn1, nn2
  integer*8, private                           :: pplan1f, pplan1b, pplan2f, pplan2b
  contains

  subroutine ft_init2(nn1_in, nn2_in)
    integer               :: nn1_in, nn2_in
    nn1 = nn1_in
    nn2 = nn2_in
    allocate(iin1(nn1), oout1(nn1))
    allocate(iin2(nn2), oout2(nn2))
    iin1 = 0.
    oout1 = 0.
    iin2 = 0.
    oout2 = 0.
    call sfftw_plan_dft_1d(pplan1f, nn1, iin1, oout1, FFTW_FORWARD, FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(pplan1b, nn1, oout1, iin1, FFTW_BACKWARD, FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(pplan2f, nn2, iin2, oout2, FFTW_FORWARD, FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(pplan2b, nn2, oout2, iin2, FFTW_BACKWARD, FFTW_ESTIMATE)
  end subroutine

  subroutine ft1axis2(adj, sign1, cdata)
    logical, intent(in)        :: adj
    integer                    :: i2
    complex, dimension(:,:) :: cdata
    real                       :: sign1
    if(adj) then
      do i2=1,nn2
        oout1(1:(nn1+1)/2) = cdata(nn1-(nn1+1)/2+1:nn1,i2)
        oout1((nn1+1)/2+1:nn1) = cdata(1:nn1-(nn1+1)/2,i2)
        call sfftw_execute_dft(pplan1b, oout1, iin1)
        cdata(:,i2) = iin1
      enddo
    else
      do i2=1,nn2
        iin1 = cdata(:,i2)
        call sfftw_execute_dft(pplan1f, iin1, oout1)
        cdata(:,i2) = oout1
        cdata(1:nn1-(nn1+1)/2,i2) = oout1((nn1+1)/2+1:nn1)
        cdata(nn1-(nn1+1)/2+1:nn1,i2) = oout1(1:(nn1+1)/2)
      enddo
    endif
    cdata = sign1*cdata/sqrt(1.*nn1)
  end subroutine

  subroutine ft2axis2(adj, sign2, cdata)
    logical, intent(in)        :: adj
    integer                    :: i1, i2
    complex, dimension(:,:) :: cdata
    real                       :: sign2
  !  write(0,*) 'inside ft2axis2', adj, sign2
  !  write(0,*) 'size cdata', size(cdata,1), size(cdata,2)
  !  write(0,*) 'nn1, nn2', nn1, nn2

    if(adj) then
      do i1=1,nn1
        oout2(1:(nn2+1)/2) = cdata(i1,nn2-(nn2+1)/2+1:nn2)
        oout2((nn2+1)/2+1:nn2) = cdata(i1,1:nn2-(nn2+1)/2)
        call sfftw_execute_dft(pplan2b, oout2, iin2)
        cdata(i1,:) = iin2
      enddo
    else
      do i1=1,nn1
        iin2 = cdata(i1,:)
        call sfftw_execute_dft(pplan2f, iin2, oout2)
!         write(0,*) 'iin2',iin2
!         write(0,*) 'oout2',oout2

        cdata(i1,1:nn2-(nn2+1)/2) = oout2((nn2+1)/2+1:nn2)
        cdata(i1,nn2-(nn2+1)/2+1:nn2) = oout2(1:(nn2+1)/2)
      enddo
    endif

!    write(0,*) 'cdata', cdata 
    cdata = sign2*cdata/sqrt(1.*nn2)
  end subroutine

end module
