program offsetmask
! Do a triangular mask around the 4th axis
! n1=z, n2=x, n3=y, n4=hx

  use sep
  use smooth_mod
  implicit none

  logical(4)          :: success, create
  character(len=128)  :: outfile1 
  integer             :: nz, nx, ny, nhx
  real                :: dz, dx, dy, dhx, oz, ox, oy
  real                :: vel, zmute, z, hx, ohx
  integer             :: iz, ix, iy, ihx

!! velocity values
  real, dimension(:,:), pointer :: wgtarray
  real, dimension(:,:,:,:), pointer :: arrayout, arrayin

!! Open I/O
  call sep_init()

  call from_param("create",create,.false.);
!! Read file dimensions (input file for masking)
  if (.not. create) then 
    call from_history("n1", nz);  call from_history("o1", oz);  call from_history("d1", dz)
    call from_history("n2", nx);  call from_history("o2", ox);  call from_history("d2", dx)
    call from_history("n3", ny);  call from_history("o3", oy);  call from_history("d3", dy)
    call from_history("n4", nhx);  call from_history("o4", ohx);  call from_history("d4", dhx)
  else
    call from_param("n1", nz);  call from_param("o1", oz);  call from_param("d1", dz)
    call from_param("n2", nx);  call from_param("o2", ox);  call from_param("d2", dx)
    call from_param("n3", ny);  call from_param("o3", oy);  call from_param("d3", dy)
    call from_param("n4", nhx);  call from_param("o4", ohx);  call from_param("d4", dhx)
  end if


!! Read linear velocity value
  call from_param("vel", vel, 1500.);
  call from_param("zmute",zmute);

  !! Read in filename
  call from_param('outfile1',outfile1);  
 

  write(0,*) '######################################'
  write(0,*) 'Program: createveltig.x'
  write(0,*) 'nz, oz, dz',  nz, oz, dz
  write(0,*) 'nx, ox, dx',  nx, ox, dx
  write(0,*) 'ny, oy, dy',  ny, oy, dy
  write(0,*) 'vel, ohx', vel, ohx
  write(0,*) '######################################'

    !! Output files
    call auxputch("n1",'i',nz,outfile1);        call auxputch("n2",'i',nx,outfile1);
    call auxputch("d1",'f',dz,outfile1);        call auxputch("d2",'f',dx,outfile1)
    call auxputch("o1",'f',oz,outfile1);        call auxputch("o2",'f',ox,outfile1)
    call auxputch("n3",'i',ny,outfile1);        call auxputch("n4",'i',nhx,outfile1);
    call auxputch("d3",'f',dy,outfile1);        call auxputch("d4",'f',dhx,outfile1)
    call auxputch("o3",'f',oy,outfile1);        call auxputch("o4",'f',ohx,outfile1)
    call auxputch("label1",'s','z(m)',outfile1);     call auxputch("label2",'s','x(m)',outfile1) 
    call auxputch("label1",'s','y(m)',outfile1);     call auxputch("label2",'s','hx(m)',outfile1) 

  call sep_close()

  allocate(arrayout(nz,nx,ny,nhx), wgtarray(nz,nhx),arrayin(nz,nx,ny,nhx))
  
  if (.not. create) then
    call sep_read(arrayin);
  else
    arrayin=1.
  end if


  do ihx=1,nhx
     do iz=1,nz
        wgtarray(iz,ihx) = 1;
     end do
  end do
  


  do ihx=1,nhx
     hx = abs(ohx + (ihx-1)*dhx)
     z = hx*vel+zmute
     iz = (z-oz)/dz+0.5
     iz = min(nz,max(1,iz));
     wgtarray(1:iz,ihx)=0.
!     write(0,*) 'iz, ihx', z,iz, ihx, hx
  end do

  wgtarray(:,1:4) = 0.
  wgtarray(:,nhx-3:nhx) = 0.

  write(0,*) 'size wgtarray', size(wgtarray,1), nz
   
  !  smooth wgtarray
  call triangle2(5,5,nz,nhx,wgtarray,wgtarray)

  write(0,*) 'after triangle2'

  do ihx=1,nhx
  do iy=1,ny
  do ix=1,nx
  do iz=1,nz
    arrayout(iz,ix,iy,ihx) = wgtarray(iz,ihx)*arrayin(iz,ix,iy,ihx)
  end do
  end do 
  end do
  end do

!! Write new vel values to output
  call sep_write(arrayout,outfile1);



  call exit(0)
   
end program 
