module parataup_mod

  use sep
  use fourier_modOMP
  use fourier_mod2OMP
  use smooth3d_mod

  implicit none
  real :: ppi=4.*atan(1.)

contains

!============================================================================

  subroutine forwardFFTinZ(inArr,outArr,cdata1,nz,oz,dz,nx,in1,out1)  

      integer   :: iz,ikz,nkz,nz,ix,nx
      real :: dz, oz, sign1, okz, dkz
      real, dimension(:,:)     :: inArr
      complex, dimension(:)  :: in1,out1
      complex, dimension(:,:)  :: outArr, cdata1

      do ix=1,nx
        do iz=1,nz
           cdata1(iz,ix) = cmplx(0.,0.)
        end do
      end do

   ! calculate wavenumber nyquist and time-domain nyquist
     nkz = nz
     okz = -1.*ppi/(dz)
     dkz = 2.*ppi/(nz*dz)
     sign1=1.

     do ix=1,nx
        do iz=1,nz
           cdata1(iz,ix) = cmplx(inArr(iz,ix))
        end do
     end do

     ! Forward 1D Fourier transform
     call ft1axis(.false., sign1, cdata1,in1,out1)

     do ix=1,nx
        do ikz=1,nkz
           outArr(ikz,ix) = cdata1(ikz,ix)
        end do
     end do


  end subroutine
!============================================================================
  subroutine inverseFFTinZ(inArr,outArr,cdata2,nz,oz,dz,nx,in1,out1)  

      integer   :: ikz,nkz, iz, nz,ix,nx
      real :: dz, oz, sign1, okz, dkz
      real, dimension(:,:)     :: outArr
      complex, dimension(:)  :: in1,out1
      complex, dimension(:,:)  :: inArr,cdata2


      do ix=1,nx
        do iz=1,nz
           cdata2(iz,ix) = cmplx(0.,0.)
        end do
      end do
     sign1=1.

   ! calculate wavenumber nyquist and time-domain nyquist
     nkz = nz
     okz = -1.*ppi/(dz)
     dkz = 2.*ppi/(nz*dz)

!    write(0,*) 'inverseFFTinZ'
!    write(0,*) 'nz, oz, dz', nz, oz, dz
!    write(0,*) 'nkz, okz, dkz', nkz, okz, dkz
!    write(0,*) 'nx', nx


     do ix=1,nx
        do ikz=1,nkz
           cdata2(ikz,ix) = inArr(ikz,ix)
        end do
     end do

 !    write(0,*) 'in 2'

     ! Forward 1D Fourier transform
     call ft1axis2(.true., sign1, cdata2,in1,out1)

 !    write(0,*) 'in 3', nx, nkz, size(cdata2,1), size(cdata2,2)
 !    write(0,*) 'outArr', size(outArr,1), size(outArr,2)

     do ix=1,nx
        do ikz=1,nkz
           outArr(ikz,ix) = real(cdata2(ikz,ix))
        end do
     end do

  !   write(0,*) 'in 4'



  end subroutine
!============================================================================

  subroutine forwardPara_oper(inArr,outArr,nkz,dkz,okz,np,dp,op,nax,dax,oax)

      integer   :: np, nax, nkz
      integer   :: ip, iax, ikz
      real      :: dp, dax, dkz
      real      :: op, oax, okz
      real      ::  p,  ax,  kz
      real      :: shift, dz
      complex      :: val, j
      complex, dimension(:,:)  :: inArr, outArr
      j = cmplx(0,1.)


    !  write(0,*) 'forwardpara'
    !  write(0,*) 'size inArr', size(inArr,1), size(inArr,2),nkz,nax
    !  write(0,*) 'size outArr', size(outArr,1), size(outArr,2),nkz,np
    
    do ip=1,np
     do ikz=1,nkz
      outArr(ikz,ip) = 0.
     end do
    end do

      do ip=1,np
         p = op+(ip-1)*dp
      do ikz=1,nkz
         kz = okz+(ikz-1)*dkz
      do iax=1,nax
         ax = oax + (iax-1)*dax
	 shift = p*ax
!!	 if (abs(shift) .gt. 2*ppi/dkz) cycle; 
	val = exp(j*kz*p*ax*ax)

!            write(0,*) 'ip,ikz,iax',ip,ikz,iax,p,kz,ax
!            write(0,*) 'j, val', j, val

        
	outArr(ikz,ip) = outArr(ikz,ip) + inArr(ikz,iax)*val
      end do
      end do
      end do


!      call erexit('exit now')

  end subroutine
 

!============================================================================

  subroutine inversePara_oper(inArr,outArr,nkz,dkz,okz,np,dp,op,nax,dax,oax)

      integer   :: np, nax, nkz
      integer   :: ip, iax, ikz
      real      :: dp, dax, dkz
      real      :: op, oax, okz
      real      ::  p,  ax,  kz
      complex      :: val, j
      complex, dimension(:,:)  :: inArr, outArr

      j = cmplx(0,1.)


      do iax=1,nax
       do ikz=1,nkz
         outArr(ikz,iax) = 0.
       end do
      end do

      do ip=1,np
         p = op+(ip-1)*dp
      do ikz=1,nkz
         kz = okz+(ikz-1)*dkz
      do iax=1,nax
         ax = oax + (iax-1)*dax
	val = exp(-1*j*kz*p*ax*ax)
	outArr(ikz,iax) = outArr(ikz,iax) +  inArr(ikz,ip)*val
      end do
      end do
      end do

  end subroutine

!============================================================================

  subroutine forwardFFTinP(inArr,outArr,cdata2,np,op,dp,nkz,in2,out2)  

      integer   :: ip,ikp,nkp,np, ikz,nkz
      real :: dp, op, okp, dkp, sign2
      complex, dimension(:)  :: in2,out2
      complex, dimension(:,:)  :: inArr, outArr,cdata2

!     write(0,*) 'inside forwardFFTinP'

   ! calculate wavenumber nyquist and time-domain nyquist
     nkp = np
     okp = -1.*ppi/(dp)
     dkp = 2.*ppi/(np*dp)
     sign2=-1.


     do ikz=1,nkz
        do ip=1,np
           cdata2(ikz,ip) = inArr(ikz,ip)
        end do
     end do

     ! Forward 1D Fourier transform
     call ft2axis2(.false., sign2, cdata2,in2,out2)

 !    write(0,*) 'cdata2', cdata2

     do ikp=1,nkp
        do ikz=1,nkz
           outArr(ikz,ikp) = cdata2(ikz,ikp)
        end do
     end do

  end subroutine

!============================================================================

  subroutine inverseFFTinP(inArr,outArr,cdata2,np,op,dp,nkz,in2,out2)  

      integer   :: ip,ikp,nkp,np, ikz,nkz
      real :: dp, op, okp, dkp, sign2
      complex, dimension(:)  :: in2,out2
      complex, dimension(:,:)  :: inArr, outArr, cdata2

   ! calculate wavenumber nyquist and time-domain nyquist
     nkp = np
     okp = -1.*ppi/(dp)
     dkp = 2.*ppi/(np*dp)
     sign2=-1.


!    write(0,*) 'inverseFFTinP'
!    write(0,*) 'np, op, dp', np, op, dp
!    write(0,*) 'nkp, okp, dkp', nkp, okp, dkp
!    write(0,*) 'nkz', nkz

     do ikz=1,nkz
        do ip=1,np
           cdata2(ikz,ip) = cmplx(0.,0.)
        end do
     end do


     do ikz=1,nkz
        do ip=1,np
           cdata2(ikz,ip) = inArr(ikz,ip)
        end do
     end do

     ! Forward 1D Fourier transform
     call ft2axis2(.true., sign2, cdata2,in2,out2)

     do ikp=1,nkp
        do ikz=1,nkz
           outArr(ikz,ikp) = cdata2(ikz,ikp)
        end do
     end do

  end subroutine

!============================================================================
  subroutine paraDecon(inoutArr,nkz,dkz,okz,nkp,dkp,okp)
 ! Deconvolution Vp(kz,kp) = Vp'(kz,kp) 1/(2ppi) sqrt(kzkp) Vp'(kz,kkp)

     integer   :: nkz, nkp, ikz, ikp
     real      :: dkz, dkp, okz, okp
     real      ::  kz,  kp, kzny, kpny, const
     complex   :: val1,val2 
     complex, dimension(:,:)   :: inoutArr

     kzny = nkz*dkz/2.
     kpny = nkp*dkp/2


     do ikp=1,nkp
        kp = abs(okp+(ikp-1)*dkp)
	
        do ikz=1,nkz
           kz=okz+(ikz-1)*dkz
	   if (kz .lt. 0) kz = kz+2*kzny
	   val1 = inoutArr(ikz,ikp)
           val2 = sqrt(kz*kp)/(2*ppi)     
           inoutArr(ikz,ikp) = val1*val2
!	   write(0,*) ikz,ikp,val,kz,kp
        end do
     end do

  end subroutine
!============================================================================
  subroutine paraDeconFinite(inoutArr,sigma,nkz,dkz,okz,nkp,dkp,okp)
 ! Deconvolution Vp(kz,kp) = Vp'(kz,kp) 1/(2ppi) sqrt(kzkp) Vp'(kz,kkp)

     integer   :: nkz, nkp, ikz, ikp
     real      :: dkz, dkp, okz, okp
     real      ::  kz,  kp, kzny, kpny
     real   :: val1,val2, eps
     complex, dimension(:,:)   :: inoutArr, sigma

     kzny = nkz*dkz/2.
     kpny = nkp*dkp/2
     eps = 1.e-9

     do ikp=1,nkp
        kp = okp+(ikp-1)*dkp
        if (kp .lt. 0) kp = kp+2*kpny
        do ikz=1,nkz
           kz=okz+(ikz-1)*dkz
	   if (kz .lt. 0) kz = kz+2*kzny
	   val1 = inoutArr(ikz,ikp)
           val2 = sqrt(kz)/(sigma(ikz,ikp)+eps)     
           inoutArr(ikz,ikp) = val1*val2
        end do
     end do

  end subroutine

!============================================================================


  subroutine sigma_init(sigma,nax,dax,oax,nkz,dkz,okz,np,dp,op,in2,out2)

    integer         :: nax, nkz, np, nkp
    integer         :: iax, ikz, ip, ikp
    real            ::  ax,  kz,  p
    real            :: dax, dkz, dp
    real            :: oax, okz, op
    complex         :: val, j
    complex,dimension(:)   :: in2,out2
    complex,dimension(:,:)   :: sigma

    nkp = np

    j= cmplx(0.,1.)

    do ikz=1,nkz
        kz = okz + (ikz-1)*dkz
    do ip=1,np
         p = op + (ip-1)*dp
       do iax=1,nax
         ax = oax+(iax-1)*dax
            val = exp(j*kz*p*ax**2)
           sigma(ikz,ip) = sigma(ikz,ip) + val
       end do
    end do
    end do

   ! FFT on 2nd axis to get sigma
   call ft2axis2(.false.,1.,sigma,in2,out2)



  end subroutine

!============================================================================

 subroutine AngleDomain_filter(n2p,npmid,op,dp,n1tau,oz,dz,input,elseoutput)

    complex, dimension(:,:) :: input,elseoutput
    complex                 :: val
    integer              :: ip, ipb,itau, n2p, n1tau,npmid
    real                 :: op, dp, oz, dz
    real, allocatable, dimension(:,:)  :: weight

    allocate(weight(n1tau,n2p))
    
    weight = 0.

        do itau=1,n1tau
             ! if (itau .gt. 0.25*n1tau) then
             !      ipb = int(((itau-0.25*n1tau)*10)/(0.75*n1tau))+3
             ! else
                   ipb = 10
             ! end if
              weight(itau,npmid+1-ipb:npmid+1+ipb) = 1.
        end do
                  
    ! smooth it a little bit to create tapering affects
!    call triangle2( 3, 3, n1tau, n2p, weight, weight)
       
    if (1>2) then
        call sep_write(weight,"wgtfile.H")
        call auxputch("n1",'i', n1tau, "wgtfile.H");      call auxputch("label1",'s', 'z', "wgtfile.H"); 
        call auxputch("d1",'f', dz, "wgtfile.H");      call auxputch("o1",'f', oz, "wgtfile.H");
        call auxputch("n2",'i', n2p, "wgtfile.H");      call auxputch("label1",'s', 'p', "wgtfile.H"); 
        call auxputch("d2",'f', dp, "wgtfile.H");      call auxputch("o1",'f', op, "wgtfile.H");
        call erexit('exit now to view wgitfile')
    end if

    do ip=1,n2p
       do itau=1,n1tau
	  val = input(itau,ip)
          input(itau,ip) = val*weight(itau,ip)
	  elseoutput(itau,ip) = val*(1.-weight(itau,ip))
       end do
    end do  

   deallocate(weight)

  end subroutine

!============================================================================

   subroutine highest(Arr,n1,n2,perc)

   
     real,dimension(:,:)   :: Arr
     real, dimension(:,:), allocatable :: tmpArr
     integer               :: n1, n2, i1, i2
     real                  :: perc, val1
     real                  :: maxv,minv,absmax


     allocate(tmpArr(n1,n2))
     maxv = maxval(Arr)
     minv = minval(Arr)
     absmax = max(abs(minv),abs(maxv))


     do i2=1,n2
     do i1=1,n1
        val1 = Arr(i1,i2)
        if (abs(val1) .gt. perc*absmax) then
          tmpArr(i1,i2) = 1.
        else
          tmpArr(i1,i2) = 0.
        end if
     end do
     end do

    call triangle2( 3, 3, n1, n2, tmpArr, tmpArr)


    do i2=1,n2
       do i1=1,n1
          Arr(i1,i2) = Arr(i1,i2)*tmpArr(i1,i2)
       end do
    end do


   end subroutine

!============================================================================

   subroutine highestComplex(Arr,n1,n2,perc)

   
     complex,dimension(:,:)   :: Arr
     real, dimension(:,:), allocatable :: AbsArr, tmpArr
     integer               :: n1, n2, i1, i2
     real                  :: perc, val1
     real                  :: maxv,minv,absmax


     allocate(tmpArr(n1,n2))
     allocate(AbsArr(n1,n2))
     do i2=1,n2
     do i1=1,n1
        AbsArr(i1,i2) = abs(Arr(i1,i2))
     end do
     end do

     maxv = maxval(AbsArr)
     minv = minval(AbsArr)
     absmax = max(abs(minv),abs(maxv))


     do i2=1,n2
     do i1=1,n1
        val1 = AbsArr(i1,i2)
        if (abs(val1) .gt. perc*absmax) then
          tmpArr(i1,i2) = 0.
        else
          tmpArr(i1,i2) = 1.
        end if
     end do
     end do

    call triangle2( 2, 2, n1, n2, tmpArr, tmpArr)


    do i2=1,n2
       do i1=1,n1
          Arr(i1,i2) = Arr(i1,i2)*tmpArr(i1,i2)
       end do
    end do


   end subroutine

!============================================================================

   subroutine subtract(comArr2,comXtalk1, n1, n2)

      complex, dimension(:,:)  :: comArr2, comXtalk1
      integer                  :: n1, n2, i1, i2
      real                     :: maxv1, maxv2, ratio1

      maxv1 = maxval(abs(comArr2))
      maxv2 = maxval(abs(comXtalk1))

      ratio1 = maxv1/maxv2

    do i2=1,n2
     do i1=1,n1
        comArr2(i1,i2) = comArr2(i1,i2)*comXtalk1(i1,i2)
     end do
    end do


   end subroutine

!============================================================================
!============================================================================

   subroutine generateMask(comArr2, n1, n2)

      complex, dimension(:,:)  :: comArr2
      real, dimension(:,:), allocatable  :: realArray
      integer                  :: n1, n2, i1, i2
      real                     :: val

      allocate(realArray(n1,n2))

    do i2=1,n2
     do i1=1,n1
	val = abs(comArr2(i1,i2))
	if (val .gt. 0) then 
	   realArray(i1,i2) = 1. 
        else
	   realArray(i1,i2) = 0. 
        end if
     end do
    end do


    call triangle2( 2, 2, n1, n2, realArray,realArray)
    
    do i2=1,n2
    do i1=1,n1

	comArr2(i1,i2) = cmplx(realArray(i1,i2),0.)
    end do
    end do

   end subroutine

!============================================================================

   subroutine generateMaskreal(realArray, n1, n2)

      real, dimension(:,:)  :: realArray

      integer                  :: n1, n2, i1, i2
      real                     :: val



    do i2=1,n2
     do i1=1,n1
	val = abs(realArray(i1,i2))
	if (abs(val) .gt. 0) then 
	   realArray(i1,i2) = 1. 
        else
	   realArray(i1,i2) = 0. 
        end if
     end do
    end do


    call triangle2( 2, 2, n1, n2, realArray,realArray)
    
   end subroutine
end module
