module rtm3d_oper

  use sep
  use omp_lib
  use type3d_mod
  use waveprop3d_mod
  use smooth3d_mod


  implicit none
  
  type(velo)       :: vel_grid

  logical, private                             :: verb, wfrout, wfmout
  logical, private                             :: smooth, VDer, lpfil, gauss, DW
  integer, private                             :: xsnap, istore
  integer, private                             :: nx, nnx, ix, ixr, bndx, irx, nrx
  integer, private                             :: ny, nny, iy, iyr, bndy, iry, nry
  integer, private                             :: nz, nnz, iz, izr, bndz, nw
  integer, private                             :: i, ii, it, nt, nrt
  integer, private                             :: nop, jsnap, nodes
  real, private                                :: dx, ox, drx, orx, idx, xtapco
  real, private                                :: dy, oy, dry, ory, idy, ytapco
  real, private                                :: dz, oz, drz, orz, idz, ztapco
  real, private                                :: dt, dt2, ot, drt, ort
  real, private                                :: zr, rho
  real, dimension(:), allocatable,private      :: xtaper, ytaper, ztaper, fs
  real, dimension(:,:,:), allocatable,private  :: vp2
  real, dimension(:,:,:), allocatable,private  :: uo, um, up, MWArray
  real, dimension(:,:,:), allocatable, private   :: DWArray
  real, dimension(:,:,:,:), allocatable,private  :: us
  character(len=128), private                  ::  wfldrec_, wfldmod_, wfldsou_, swin_
  integer, dimension(:), allocatable, private  :: UPmarker, RECmarker
  real                                         :: dtsav,  otsav
  integer                                      :: ntsav, imark, irt, ifile, bfile, efile

  ! Expanding Domain Stuff
  integer :: nzmax
  real    :: t, vmax

  ! Blocky stuff
  integer ::  zbsz, xbsz, ybsz
  
  contains 

  integer function L_init( sougrid_in, recgrid_in, velgrid_in, us_in, vp_in, rho_in, bc_in, &
  		wfldrec_in, wfldmod_in, jsnap_in, verb_in, MWfile_in, DWfile_in,nodes_in) result(stat)


    optional :: wfldrec_in, wfldmod_in, jsnap_in, verb_in, DWfile_in

    type(sou_grid)       :: sougrid_in
    type(rec_grid)       :: recgrid_in
    type(velo)           :: velgrid_in
    type(bc_values)     :: bc_in

    logical              :: verb_in
    integer              :: jsnap_in, nodes_in
    real                 :: rho_in
    character(len=128)   :: wfldrec_in, wfldmod_in, us_in, vp_in
    character(len=128)   :: MWfile_in, DWfile_in


    nx = velgrid_in%x%n
    dx = velgrid_in%x%d
    ox = velgrid_in%x%o
    ny = velgrid_in%y%n
    dy = velgrid_in%y%d
    oy = velgrid_in%y%o
    nz = velgrid_in%z%n
    dz = velgrid_in%z%d
    oz = velgrid_in%z%o
    smooth = velgrid_in%smooth
    
    ntsav = sougrid_in%ntsav
    bfile = sougrid_in%bfile
    efile = sougrid_in%efile
    
    nrx = recgrid_in%x%n
    orx = recgrid_in%x%o
    drx = recgrid_in%x%d
    nry = recgrid_in%y%n
    ory = recgrid_in%y%o
    dry = recgrid_in%y%d
    zr = recgrid_in%z%o

    nrt = recgrid_in%t%n
    drt = recgrid_in%t%d
    ort = recgrid_in%t%o

    nt = recgrid_in%nt
    dt = recgrid_in%dt
    ot = recgrid_in%ot
   
    bndx = bc_in%bndx
    bndy = bc_in%bndy
    bndz = bc_in%bndz
    
    nop = bc_in%nop
    nnx = nx - 2*bndx
    nny = ny - 2*bndy
    nnz = nz - 2*bndz

    nw = (velgrid_in%zw - oz)/dz + .5
    wfrout  = bc_in%wfrout
    wfmout   = bc_in%wfmout
    DW       = bc_in%DW
    lpfil    = bc_in%lpfil
    VDer     = bc_in%VDer
    gauss    = bc_in%gauss




    xsnap = sougrid_in%t%d/recgrid_in%t%d + 0.5
    jsnap = jsnap_in

    xbsz=40; ybsz=40; zbsz=40;
    vmax = 4500.
    nodes = nodes_in

! ********************
    write(0,*) 'xsnap', xsnap, sougrid_in%t%d, recgrid_in%t%d
    write(0,*) 'jsnap', jsnap
! ********************
    allocate(UPmarker(ntsav), RECmarker(nrt))
    allocate(vp2(nz,nx,ny)) 
	
    ! Load up the RECmarker
    do i=1,nrt
       t = ort + (i-1)*drt
       ii = (t-ot)/dt + 1.5
       RECmarker(i) = ii
    !   write(0,*) 'RECmarker', i, ii
    end do


    do iy=1,ny
      do ix=1,nx
        do iz=1,nz
           vp2(iz,ix,iy) = 0.
        end do
      end do
    end do

    if (wfrout) then
       wfldrec_ = wfldrec_in
    end if
    if (wfmout) then
       wfldmod_ = wfldmod_in
    end if
    wfldsou_ = us_in

    if (present(verb_in)) verb = verb_in

  
   ! Read in the background velocity
   call sep_read( vp2(1+bndz:nz-bndz,1+bndx:nx-bndx,1+bndy:ny-bndy), vp_in, esize=4)

   ! padding
   do iz=1,bndz; vp2(iz,:,:) = vp2(bndz+1,:,:); end do 
   do iz=1,bndz; vp2(nz-iz+1,:,:) = vp2(nz-bndz,:,:); end do 
   do ix=1,bndx; vp2(:,ix,:) = vp2(:,1+bndx,:); vp2(:,nx-ix+1,:) = vp2(:,nx-bndx,:); end do 
   do iy=1,bndy; vp2(:,:,iy) = vp2(:,:,1+bndy); vp2(:,:,ny-iy+1) = vp2(:,:,ny-bndy); end do 


    write(0,*) 'allocating 3X', nz,nx,ny, nz*nx*ny
    allocate(up(nz,nx,ny))
    allocate(um(nz,nx,ny))
    allocate(uo(nz,nx,ny))
        
    !----------------------------------------------------------------
    !!Tapersetup  
    allocate(ztaper(bndz))
    allocate(xtaper(bndx), ytaper(bndy))

    
    ztaper=0.; xtaper=0.; ytaper=0.;


    !----------------------------------------------------------------

    dt2 = dt*dt
    idx = dt2/dx**2; idy = dt2/dy**2; idz = dt2/dz**2;
    write(0,*) 'idx, idy, idz', idx, idy, idz
    write(0,*) 'dt,dx, dy, dz', dt, dx, dy, dz

    call  tapersetupCos3d(bndz,bndx,bndy,ztaper,xtaper,ytaper,0.98,0.02)


    do iy=1,ny
    do ix=1,nx
       do iz=1,nz      
          vp2(iz,ix,iy) = vp2(iz,ix,iy)*vp2(iz,ix,iy) 
       end do
    end do
    end do

    write(0,*) 'max vel', maxval(vp2)


     ! Smoothing velocity model
     if (smooth) then
      write(0,*) 'smoothing the velocity'
      call triangle3(8, 8, 8, nz, nx, ny, vp2, vp2)
     end if

     
    write(0,*) 'pass 6'

    allocate(DWArray(nrx,nry,nrt))
    do it=1,nrt
      do iry=1,nry
        do irx=1,nrx
           DWArray(irx,iry,it) = 1.
        end do
      end do
    end do

    if (DW) then        
        write(0,*) 'DWfile size', size(DWArray,1), size(DWArray,2), size(DWArray,3)
        write(0,*) 'DWfile name', DWfile_in
        call sep_read(DWArray,DWfile_in)
    end if


    ! Model Weighting
    allocate(MWArray(nnz-nw,nnx,nny))
    write(0,*) 'MWfile', MWfile_in
    call sep_read(MWArray,MWfile_in, dim1=nnz-nw, dim2=nnx, dim3=nny, esize=4)

    write(0,*) 'pass 7'
 

  if (1 == 2) then
         call sep_write(MWArray,'MWarray.H',dim1=nnz-nw, dim2=nnx,dim3=nny,esize=4) 
 	 call auxputch("n1",'i',nnz-nw,'MWarray.H');   
         call auxputch("d1",'f',dz,'MWarray.H');   call auxputch("o1",'f',oz,'MWarray.H');
	 call auxputch("label1", 's', 'Z', 'MWarray.H');
         call auxputch("n2",'i',nnx,'MWarray.H');   
         call auxputch("d2",'f',dx,'MWarray.H');   call auxputch("o2",'f',ox,'MWarray.H');
	 call auxputch("label2", 's', 'X', 'MWarray.H');
         call auxputch("n3",'i',nny,'MWarray.H');   
         call auxputch("d3",'f',dy,'MWarray.H');   call auxputch("o3",'f',oy,'MWarray.H');
	 call auxputch("label3", 's', 'Y', 'MWarray.H');
	 call erexit('exit now')
   end if


    ! Allocation
    write(0,*) 'allocating us', (nnz-nw), nnx, nny, ntsav
    write(0,*) 'allocating us', (nnz-nw)*nnx*nny*ntsav
    allocate(us(nnz-nw,nnx,nny,ntsav))
    
    do it=1,ntsav
      do iy=1,nny
        do ix=1,nnx
          do iz=1,nnz-nw
             us(iz,ix,iy,it)=0.
          end do
        end do
      end do
    end do

    if ( 1 == 2 ) then
         write(0,*) '%%% writing to velout', "velout.H"
         call auxputch("n1",'i',nz,"velout.H");
         call auxputch("d1",'f',25.,"velout.H");   call auxputch("o1",'f',0.,"velout.H");
         call auxputch("n2",'i',nx,"velout.H");
         call auxputch("d2",'f',25.,"velout.H");   call auxputch("o2",'f',50000.,"velout.H");
         call auxputch("n3",'i',ny,"velout.H");
         call auxputch("d3",'f',25.,"velout.H");   call auxputch("o3",'f',34450.,"velout.H");
         write(0,*) 'size vp2', size(vp2)
         call sep_write(vp2, "velout.H")
!         call sep_write(vp2, "velout.H", dim1=nz, dim2=nx, dim3=ny, esize=4)
         call erexit('eeexit now')
    end if



   ! --------------------------------------------------------------
 
    stat=0

  end function L_init

  !---------------------------------------------------------------------------
  !***************************************************************************
  !---------------------------------------------------------------------------
  integer function L_oper(adj, add, model, data) result(stat)


    logical, intent(in)         :: adj, add
    real, dimension(:,:,:)  :: model, data 
    integer                   :: ix, iy, iz

    stat = 1

    if (.not. add) then
       if (adj) then
          do iy=1,nny
            do ix=1,nnx
              do iz=1,nnz-nw
                 model(iz,ix,iy) = 0.
              end do
            end do
          end do
       else
          do irt=1,nrt
            do iry=1,nry
              do irx=1,nrx
                data(irx,iry,irt)=0.
              end do
            end do
          end do
       end if
    end if

    izr = (zr-oz)/dz + 1.5 + bndz

 !   write(0,*) 'zr, izr, bndz', zr, izr, bndz 


    ! initializing the array
!!    !$OMP parallel do default(shared) schedule(dynamic,1) num_threads(nodes) private(ix,iy,iz)
    do iy=1,ny
    do ix=1,nx
       do iz=1,nz
          up(iz,ix,iy)=0.; um(iz,ix,iy)=0.; uo(iz,ix,iy)=0.;
       end do          
    end do
    end do
!!    !$OMP end parallel do
    
    if (adj) then      !!! ### RTM ### !!!


      ! Apply weighting matrix to data space       
      do iry=1,nry
        do irx=1,nrx
          do irt=1,nrt
             data(irx,iry,irt) = data(irx,iry,irt)*DWArray(irx,iry,irt)
          end do
        end do
      end do


      imark = 0
      irt = nrt
      ifile = efile


      do it=nt,1,-1
         

         if ( (imark .le. 0) .and. (ifile .ge. bfile) ) then
            call getfilename(wfldsou_,swin_,ifile);
            call from_aux(swin_, "n4", ntsav);  
            call from_aux(swin_,"d4", dtsav);   
            call from_aux(swin_,"o4", otsav);

            ! Read in the sou wavefield
            write(0,*) 'reading file', swin_, it
            call sep_read( us, swin_, esize=4)

            ! Load up the UPmarker
            do i=1,ntsav
               t = otsav + (i-1)*dtsav
               ii = (t-ot)/dt + 1.5
               UPmarker(i) = ii
      !       write(0,*) 'UPmarker', i, ii
            end do

            imark = ntsav
            ifile = ifile-1
       end if


            !! Expanding grid calculation
            !t = (it-1)*dt+ot
            nzmax = int(max(real(zbsz),min(((vmax*t-oz)/dz + bndz+3*nop + 1),real(nz))))
            !nzmax = nz


           !! 16th order laplacian 
           call laplac3d16(adj, uo,up, vp2, nx, ny, nz, nzmax, xbsz,ybsz,zbsz, idx,idy,idz, nop,it,nodes)


           !! Inject receiver signal  
           if ((irt .ge. 1) .and. (irt .le. nrt) .and. (it .eq. RECmarker(irt))) then
!!            !$OMP parallel do default(shared) schedule(dynamic,1) num_threads(nodes) private(irx,ixr,iry,iyr)
              do iry=1,nry  
                 iyr = (ory+(iry-1)*dry -oy)/dy + 1.5 + bndy
                 do irx=1,nrx
                    ixr = (orx+(irx-1)*drx -ox)/dx + 1.5 + bndx
                    up(izr,ixr,iyr) = up(izr,ixr,iyr) + data(irx,iry,irt)  
                 end do
               end do 
!!          !$OMP end parallel do
             irt = irt-1
          end if


           !! time step
!!           !$OMP parallel do default(shared) schedule(dynamic,1) num_threads(nodes) private(ix,iy,iz)
           do iy=1,ny
           do ix=1,nx
              do iz=1,nz
                 up(iz,ix,iy)= up(iz,ix,iy)-um(iz,ix,iy)
                 um(iz,ix,iy)=uo(iz,ix,iy); 
                 uo(iz,ix,iy)=up(iz,ix,iy);
                 up(iz,ix,iy)=0.
              end do          
           end do
           end do
!!          !$OMP end parallel do
          

          !! Taper
  	  call taperingCos3d(nz,nx,ny,bndz,bndx,bndy,ztaper,xtaper,ytaper,uo)
	  call taperingCos3d(nz,nx,ny,bndz,bndx,bndy,ztaper,xtaper,ytaper,up)
	  call taperingCos3d(nz,nx,ny,bndz,bndx,bndy,ztaper,xtaper,ytaper,um)


         ! if ((mod(it,100) .eq. 0)) write(0,*) 'adjoint it', it, sum(uo), sum(model(:,:,:))

         !!! RTM: IMAGING CONDITION
         if (imark .gt. 0) then
           if ( it .eq. UPmarker(imark)  ) then	        
!!	   !$OMP parallel do default(shared) schedule(dynamic,1) num_threads(nodes) private(ix, iy, iz)
             do iy=1,nny
               do ix=1,nnx
                 do iz=1,(nnz-nw)
	            model(iz,ix,iy) = model(iz,ix,iy) + us(iz,ix,iy,imark)*uo(iz+bndz+nw,ix+bndx,iy+bndy)
                 enddo
               enddo
             enddo
!!         !$OMP end parallel do
!	     write(0,*) 'after IC', it, sum(us(:,:,:,imark)), sum(uo), sum(model)
             imark = imark-1            
           end if
         end if       
         

         !! output receiver movie if optioned                 
         if (mod(it,jsnap)==0 .and. wfrout ) then
            write(0,*) 'outputting rec movie,it', it, nz, nx, ny         
            call sep_write(uo(1:nz,1:nx,1:ny),wfldrec_,dim1=nz,dim2=nx,dim3=ny,esize=4)
         end if             

         !! output model movie if optioned  
         if(mod(it,jsnap)==0 .and. wfmout  ) then
            istore = it/jsnap
            write(0,*) 'outputting model movie', it
            call sep_write(model(1:nnz-nw,1:nnx,1:nny), wfldmod_, dim1=nnz-nw, dim2=nnx, dim3=nny, esize=4)       
         end if

   enddo ! (end rec time loop)
     


    ! Apply filter to model
    if (gauss) then
	 call Gaussian3D(adj,model,nnz-nw,nnx,nny)
   else if (lpfil) then
	 call  Laplacian3D(adj,model,nnz-nw,nnx,nny)
    else if (VDer) then
        call Ver2ndDer3D(adj,model,nnz-nw,nnx,nny)
    end if


     ! Apply weighting to model
     do iy=1,nny   
       do ix=1,nnx
         do iz=1,(nnz-nw)
           model(iz,ix,iy) = model(iz,ix,iy)*MWArray(iz,ix,iy)
         end do
       end do
     end do

 

   else  !!! ### FORWARD Modelling ### !!!

      imark = ntsav+1
      irt = 1
      ifile = bfile

     ! Apply weighting to model
     do iy=1,nny   
       do ix=1,nnx
         do iz=1,(nnz-nw)
           model(iz,ix,iy) = model(iz,ix,iy)*MWArray(iz,ix,iy)
         end do
       end do
     end do
     
    ! Apply filter to model
    if (gauss) then
	 call Gaussian3D(adj,model,nnz-nw,nnx,nny)
   else if (lpfil) then
	 call  Laplacian3D(adj,model,nnz-nw,nnx,nny)
    else if (VDer) then
        call Ver2ndDer3D(adj,model,nnz-nw,nnx,nny)
    end if

         write(0,*) '*****************************'

          write(0,*) 'size data', size(data,1), size(data,2), size(data,3)
          write(0,*) 'size uo', size(uo,1) ,size(uo,2), size(uo,3)

          write(0,*) 'ory, dry, nry', ory, dry, nry
          write(0,*) 'oy, dy, ny', oy, dy, ny, bndy

         write(0,*) '*****************************'

           
     do it=1,nt

       if ( (imark .ge. ntsav+1) .and. (ifile .le. efile) ) then
          call getfilename(wfldsou_,swin_,ifile);
          call from_aux(swin_, "n4", ntsav);  
          call from_aux(swin_,"d4", dtsav);   
          call from_aux(swin_,"o4", otsav);

          ! Read in the sou wavefield
            write(0,*) 'reading file', swin_, it
          call sep_read( us, swin_, esize=4)

          ! Load up the UPmarker
          do i=1,ntsav
             t = otsav + (i-1)*dtsav
             ii = (t-ot)/dt + 1.5
             UPmarker(i) = ii
       !    write(0,*) 'UPmarker', i, ii
          end do

          imark = 1
          ifile = ifile+1

       end if


        !! Taper
	call taperingCos3d(nz,nx,ny,bndz,bndx,bndy,ztaper,xtaper,ytaper,uo)
	call taperingCos3d(nz,nx,ny,bndz,bndx,bndy,ztaper,xtaper,ytaper,up)
	call taperingCos3d(nz,nx,ny,bndz,bndx,bndy,ztaper,xtaper,ytaper,um)

                  
        !! Expanding grid calculation
        ! t = (it-1)*dt+ot
         nzmax = int(max(real(zbsz),min(((vmax*t-oz)/dz + bndz+3*nop + 1),real(nz))))
        !nzmax = nz


        ! if (mod(it,20) .eq. 1) write(0,*) 'it', it, sum(uo), sum(up), sum(um)

        !! 16th order laplacian 
        call laplac3d16(adj, uo,up, vp2, nx, ny, nz, nzmax, xbsz,ybsz,zbsz, idx,idy,idz, nop,it,nodes)


        ! if (mod(it,20) .eq. 1) write(0,*) '@it', it, sum(uo), sum(up), sum(um)

        ! Forward modelling Injection
        if (imark .le. ntsav) then
	  if (  it .eq. UPmarker(imark)  ) then
          ! write(0,*) 'inject scattered wave', it, imark, sum(us(:,:,:,imark)), sum(model)
!!        !$OMP parallel do default(shared) schedule(dynamic,1) num_threads(nodes) private(ix, iz, iy)                   
             do iy=1,nny        
               do ix=1,nnx
                 do iz=1,(nnz-nw)
                   up(iz+nw+bndz,ix+bndx,iy+bndy) = up(iz+nw+bndz,ix+bndx,iy+bndy) &
                          + us(iz,ix,iy,imark)*model(iz,ix,iy) 

                 enddo
               enddo
             end do
!!        !$OMP end parallel do
             imark = imark+1
           end if
        end if

       ! if (mod(it,20) .eq. 1) write(0,*) '@@it', it, sum(uo), sum(up), sum(um)


           !! time step
!!           !$OMP parallel do default(shared) schedule(dynamic,1) num_threads(nodes) private(ix,iy,iz)
           do iy=1,ny
           do ix=1,nx
              do iz=1,nz
                 up(iz,ix,iy)=up(iz,ix,iy)-um(iz,ix,iy)
                 um(iz,ix,iy)=uo(iz,ix,iy); 
                 uo(iz,ix,iy)=up(iz,ix,iy);
                 up(iz,ix,iy)=0.
              end do          
           end do
           end do
!!          !$OMP end parallel do


          !! Collect data
          if ((irt .le. nrt) .and. (it .eq. RECmarker(irt))) then
!!!  !!	  !$OMP parallel do default(shared) schedule(dynamic,1) num_threads(nodes) private(irx,ixr,iry,iyr)
             do iry=1,nry
                iyr = (ory+(iry-1)*dry -oy)/dy + 1.5 + bndy
               do irx=1,nrx
                  ixr = (orx+(irx-1)*drx -ox)/dx + 1.5 + bndx

          !         write(0,*) 'irx,ixr,iry,iyr',irx,ixr,iry,iyr,izr
 

                  data(irx,iry,irt) = data(irx,iry,irt)  + uo(izr,ixr,iyr)
               end do
             end do
!!! !!       !$OMP end parallel do
            irt = irt+1
         end if

      !   call erexit('exit now to check receiver dimensions')


          if ((mod(it,500) .eq. 1)) write(0,*) 'forward sum, it', it,irt, sum(data(:,:,irt-1)), sum(uo),sum(up),sum(um)

          if (mod(it,jsnap)==0 .and. wfrout .and. (it .ne. nt) ) then
      	     write(0,*) 'output ur', it, sum(uo)
             call sep_write(uo(1:nz,1:nx,1:ny),wfldrec_,dim1=nz,dim2=nx,dim3=ny,esize=4)
          end if
                                                   
     end do ! (rec time loop)
          
   ! Apply weighting matrix to data space
     do iry=1,nry
       do irx=1,nrx
         do irt=1,nrt
            data(irx,iry,irt) = data(irx,iry,irt)*DWArray(irx,iry,irt)
         end do
       end do
     end do

      
   end if




   ! ------------------------------------------------------


    stat =0

    end function

    !---------------------------------------------------------------------------
    !---------------------------------------------------------------------------
    integer function L_close() result(stat)

      stat=1

	write(0,*) 'dalloc 1'
      deallocate(vp2)
      	write(0,*) 'dalloc 2'
      deallocate(xtaper, ztaper, ytaper)
	write(0,*) 'dalloc 3'
      deallocate(um, up, uo)
	write(0,*) 'dalloc 4'
      deallocate(us)
        write(0,*) 'dalloc 5'
      deallocate(UPmarker, MWArray, DWArray)

      
      stat=0

    end function L_close

end module
