#! /usr/bin/env python
#from SEP import *
import commands as cmd
import re,string
import sys, os
import os.path as path



from pyBath import *

import pyInvmodRegJoint
import pyfirstiterRegJoint


# ----------------------------------------------------------------
# ----- Project specific Definition -------- *** Need Changing ***
# ----------------------------------------------------------------

# Please make sure pyAdjoint and pyForward include the proper pyBath.py file that contains the location of the nodes
# L = [[0],['A',GxA,GyA,GzA],['B',GxB,GyB,GzB],['C',GxC,GyC,GzC], ...]
sbin = "./Bin/"
maindir = "."
projdir = maindir + "/LSM/J"            # Project directory, where all intermediate output stored
plet = 'J'                                  # Project letter (to avoid overwrite in scr)


# ---- File names and project directories ----------------
velfile= 'Input/velDeimos3D.H'        # vel file (located at projdir/Input/.)
parfile = 'PMigSddA.par'              # Adjoint param file (located at projdir/Par/.)
dirdata1 = 'fiedat/D/fdDeiD'
dirdata2 = 'fiedat/A/fdDeiA'

souwfld1 =  'souwfld/B/swDeiB' 
souwfld2 =  'souwfld/A/swDeiA' 

DWfile1  = 'DWfile/H/DWDeiHb'       # Data weighting files - up
MWfile1  =  'MWfile/E/mwDeiE'         # Model weighting files - up

DWfile2  = 'DWfile/A/DeiMskA'       # Data weighting files - down
MWfile2  =  'MWfile/E/mwDeiE'         # Model weighting files - down


# ---------- Preconditioning and Regularization -----------------------
Rname = 'DSO3d.x'                           # Program name for regularization operation
regParm ='Par/Pdso.par'
regax =4
eps   = 2.00                               # Epsilon is 20% between ratio of initial ggmDggm and ggDgg
ratio = 0.                                 # Initialization for ggdDggd/ggmDggm
smooth = 0                                    
lpfil  = 0
gauss  = 0
VDer   = 1
ampba  = 1                                 # [0]=none, 1=Joint, 2=down, 3=up
DW     = 1                                 # Data space weighting

# ---- Image space parameters ---------------
modax = 4                                  # number of axis in image file
nmz = 287                                  # nz 
omz = 850.                                 # oz
dmz = 25.                                  # dz
nmx = 420                                  # nx
omx = 50000.                               # ox
dmx = 25.                                  # dx
nmy = 28                                   # ny
omy = 34450.                               # oy
dmy = 25.                                  # dy
nhx = 30

# --- Data space parameters ----------------
datax = 3                               # number of axis in data space

# ----- Iteration loop ----------------------
biter = 1	                             # Beginning iteration
eiter = 25                             # End iteration


# ----------------------------------------------------------------
# ------- End of Definition --------------------------------------
# ----------------------------------------------------------------

dirm  = projdir+'/m/mDei'+plet
dirs  = projdir+'/s/sDei'+plet
dirdm = projdir+'/dm/dmDei'+plet
dirgg = projdir+'/gg/ggDei'+plet
dirss = projdir+'/ss/ssDei'+plet
dirrr = projdir+'/rr/rrDei'+plet 

dirggm = projdir+'/ggm/ggmDei'+plet
dirssm = projdir+'/ssm/ssmDei'+plet
dirrrm = projdir+'/rrm/rrmDei'+plet
dmm    = projdir + '/dm/dmMDei'+plet

os.system("cd ..")
compie=1
if (compie == 1):
	os.system("make Bin/modcgstep.x")
	os.system("make Bin/Mig_Main3D.x")
	os.system("make Bin/Forward3D.x")
	os.system("make Bin/updateAB.x")
	os.system("make Bin/Solver_ops.x")
	os.system("make Bin/ApplyWgt3D.x")
	os.system("make Bin/SUMMING.x")

	cmd1= 'make Bin/%s'%(Rname)
	os.system(cmd1)

refiter=1

if (biter == 1):
	pyfirstiterRegJoint.run(maindir,projdir,velfile,parfile, plet,K, nmz,omz,dmz,nmx,omx,dmx, nmy,omy,dmy,nhx, modax,datax,dirdata1, dirdata2, smooth,lpfil,VDer, gauss, DW, souwfld1, souwfld2,MWfile1, MWfile2, DWfile1, DWfile2, Rname,regParm, regax, eps,ratio)
	biter = biter+1



# At this point, iter 1 is already finished, so calculate ratio
ratio = pyInvmodRegJoint.Ratio(maindir,projdir,plet,K,Rname,regParm, modax,datax,1)
print 'ggmDggm vs ggdDggd ratio', ratio

for iter in range(biter,eiter+1):

	print 'Running iteration#', iter


	# Step 2: Loop shots  dm = F'Rd  + eps A' Rm
	if (iter != -1*biter):
		pyInvmodRegJoint.Adjoint(maindir, projdir, velfile, parfile, plet, smooth, lpfil, VDer, gauss, DW, souwfld1, souwfld2, MWfile1, MWfile2, DWfile1, DWfile2, Rname, regParm, eps,ratio, modax, nhx, iter)
		pyInvmodRegJoint.Add(maindir,projdir,dirdm, iter, modax, plet,dmm,dirdm)
		C1="%s/offsetmask.x < %sSUMiter%s.H vel=6. zmute=100. outfile1=%sTotaliter%s.H >/dev/null"%(sbin,dirdm,iter,dirdm,iter)
		C2= 'echo "label1=z label2=x label3=y label4=hx" >> %sTotaliter%s.H '%(dirdm,iter)
		print C1
		print C2
		os.system(C1)
		os.system(C2)

	# Step 4: Loop shots  gg = Fdm and ggm= eps A dm        
	if (iter != -1*biter):
		pyInvmodRegJoint.Forward(maindir, projdir, velfile, parfile, plet, smooth, lpfil, VDer, gauss, DW, souwfld1, souwfld2, MWfile1, MWfile2, DWfile1, DWfile2, Rname, regParm, eps,ratio, modax, nhx,iter)
			        
       
       
        # Step 5: Calculate the alpha and beta in conjugate direction
        alpha, beta = pyInvmodRegJoint.DPcalCD(projdir, plet, iter,datax,modax, dirdm, dirggm, dirgg, dirrr, dirrrm, dirss, dirssm)
	print alpha, beta


	#  Step 6: Update the residual files
	print 'update data residual, iter', iter
	for i in K[:]:
        	print 'residual update, iter', iter, 'shot', i
                #  calculate your ss file ss=alpha*dr+beta*ss	
        	C1 = "Bin/updateAB.x alpha=%s beta=%s naxis=%s file1=%sup%siter%s.H file2=%sup%siter%s.H outfile=%sup%siter%s.H >/dev/null"%(alpha, beta, datax, dirgg, i, iter, dirss, i, iter, dirss, i, iter+1)
		print C1
		os.system(C1)
		C1 = "Bin/updateAB.x alpha=%s beta=%s naxis=%s file1=%sdn%siter%s.H file2=%sdn%siter%s.H outfile=%sdn%siter%s.H >/dev/null"%(alpha, beta, datax, dirgg, i, iter, dirss, i, iter, dirss, i, iter+1)
		print C1
		os.system(C1)
			             
		#  update your current residual here  rr=rr+ss
		C1 = "Bin/updateAB.x alpha=1.0 beta=1.0 naxis=%s file1=%sup%siter%s.H file2=%sup%siter%s.H outfile=%sup%siter%s.H >/dev/null"%(datax, dirrr, i, iter, dirss, i, iter+1, dirrr, i, iter+1)	
		print C1
		os.system(C1)
		C1 = "Bin/updateAB.x alpha=1.0 beta=1.0 naxis=%s file1=%sdn%siter%s.H file2=%sdn%siter%s.H outfile=%sdn%siter%s.H >/dev/null"%(datax, dirrr, i, iter, dirss, i, iter+1, dirrr, i, iter+1)	
		print C1
		os.system(C1)


	#Step 6b: Update the  model residual files
	print 'update model residual, iter', iter
	# calculate ssm = alpha*ggm 
	C1 = "Bin/updateAB.x alpha=%s cg=1 naxis=%s file1=%siter%s.H outfile=%siter%s.H >/dev/null"%(alpha,modax,dirggm,iter,dirssm,iter+1)
	print C1	
	os.system(C1)  
	# Calculate your new rrm = rrm + ssm
	C1 = "Bin/updateAB.x alpha=1.0 beta=1.0 naxis=%s file1=%siter%s.H file2=%siter%s.H outfile=%siter%s.H >/dev/null"%(modax,dirrrm,iter,dirssm,iter+1,dirrrm,iter+1)	
	print C1
	os.system(C1)

	
	# Step 7: Update the model files
	# Calculate your new dm s=alpha*dm+beta*s
	print 'model update, iter', iter
	C1 = "Bin/updateAB.x alpha=%s beta=%s naxis=%s file1=%sTotaliter%s.H file2=%sTotaliter%s.H outfile=%sTotaliter%s.H >/dev/null"%(alpha,beta,modax,dirdm,iter,dirs,iter,dirs,iter+1)	
	print C1
	os.system(C1)  
	# Calculate your new m
	C1 = "Bin/updateAB.x alpha=1.0 beta=1.0 naxis=%s file1=%sTotaliter%s.H file2=%sTotaliter%s.H outfile=%sTotaliter%s.H >/dev/null"%(modax,dirm,iter,dirs,iter+1,dirm,iter+1)	
	print C1
	os.system(C1)
       
#        # Step 9, remove files
#	if (iter > 2):
#		pyInvmodRegJoint.removefiles(iter,maindir, plet, rmlist)


# Calculate final residuals and save to report
resid = pyInvmodRegJoint.ResidCal(projdir, plet,K,iter+1,datax,dirrr)
print  resid
file = open(projdir+'/Reportsave.txt', 'a')              # save to report
cmd1 = " %s na na %s \n"%(iter, resid)
file.writelines(cmd1);
file.close()
